from collections import OrderedDict
from math import inf
from types import FunctionType
from warnings import warn
from . import get_simulator
from .simulation import Simulator


class Environment(OrderedDict):
    """ This class represents the exchange of compounds between an organism and the environment. """

    def __init__(self):
        OrderedDict.__init__(self)

    def __str__(self):
        entries = (f"{r_id}\t{lb}\t{ub}" for r_id, (lb, ub) in self.items())
        return '\n'.join(entries)

    def __repr__(self):
        return str(self)

    def get_compounds(self, fmt_func=None):
        """
        Return the list of compounds in the growth medium for this environment.
        Args:
            fmt_func (str or function): python format string (see Notes)
        Returns:
            list: compounds in the medium
        Notes:
            The format function parameter is used to convert exchange reaction ids to metabolite ids.
            By default assumes BiGG notation ('R_EX_h2o_e' is transformed to 'h2o').
        """

        if fmt_func is None:
            def fmt_func(x):
                if x[:2] == 'R_':
                    return x[5:-2]
                else:
                    return x[3:-2]
        elif not isinstance(fmt_func, FunctionType):
            raise RuntimeError("fmt_func argument must be a string or function.")

        compounds = []

        for r_id, (lb, _) in self.items():
            if lb < 0:
                compounds.append(fmt_func(r_id))

        return compounds

    def apply(self, model, exclusive=True, inplace=True, warning=True, prefix=None):
        """
        Apply environmental conditions to a given model
        Args:
            model (REFRAME or COBRAPY model or Simulator): model
            exclusive (bool): block uptake of any model compounds not specified in this environment (default: True)
            warning (bool): print warning for exchange reactions not found in the model (default: True)
            inplace (bool): apply to model, otherwise return a constraints dict (default: True)
        """
        if not isinstance(model, Simulator):
            sim = get_simulator(model)
        else:
            sim = model

        if prefix:
            e = {f"{prefix}{rxn_id}": (lb, ub) for rxn_id, (lb, ub) in self.items()}
        else:
            e = self

        if exclusive:
            env = Environment.empty(model)
            env.update(e)
        else:
            env = e

        if not inplace:
            constraints = {}

        for r_id, (lb, ub) in env.items():
            if r_id in sim.reactions:
                if inplace:
                    sim.set_reaction_bounds(r_id, lb, ub)
                else:
                    constraints[r_id] = (lb, ub)
                    
        if not inplace:
            return constraints

    def simplify(self, inplace=False):
        """ Keep only uptake reactions for the respective medium. """

        if inplace:
            env = self
        else:
            env = self.copy()

        to_remove = []

        for r_id, (lb, _) in env.items():
            if lb >= 0:
                to_remove.append(r_id)

        for r_id in to_remove:
            del env[r_id]

        if not inplace:
            return env

    @staticmethod
    def from_reactions(reactions, max_uptake=10.0):
        """
        Create an environment from list of uptake reactions
        Arguments:
            reactions (list): exchange reactions
            max_uptake (float): maximum uptake rate for given compounds (default: 10.0)
        Returns:
            Environment
        """

        env = Environment()
        for r_id in reactions:
            env[r_id] = (-max_uptake, inf)

        return env

    @staticmethod
    def from_compounds(compounds, fmt_func=None, max_uptake=10.0, prefix=''):
        """
        Initialize environment from list of medium compounds
        Arguments:
            compounds (list): List of compounds present in the medium
            fmt_func (str or function): python format string (see Notes)
            max_uptake (float): maximum uptake rate for given compounds (default: 10.0)
        Returns:
            Environment
        Notes:
            Please provide a formatting string or function to convert compound identifiers to exchange reactions.
            Default format uses BiGG notation: "R_EX_{}_e" (for example "h2o" becomes "R_EX_h2o_e").
        """

        if fmt_func is None:
            def fmt_func(x):
                return f"{prefix}EX_{x}_e"
        elif isinstance(fmt_func, str):
            fmt_str = fmt_func
            def fmt_func(x): return fmt_str.format(x)
        elif not isinstance(fmt_func, FunctionType):
            raise RuntimeError("fmt_func argument must be a string or function.")

        reactions = map(fmt_func, compounds)

        return Environment.from_reactions(reactions, max_uptake=max_uptake)

    @staticmethod
    def from_model(model):
        """
        Extract environmental conditions from a given model
        Arguments:
            model (REFRAME or COBRAPY model or Simulator): model
        Returns:
            Environment: environment from provided model
        """

        if not isinstance(model, Simulator):
            sim = get_simulator(model)
        else:
            sim = model

        env = Environment()

        for r_id in sim.get_exchange_reactions():
            env[r_id] = sim.get_reaction_bounds(r_id)

        return env

    @staticmethod
    def from_defaults(model, max_uptake=10.0, max_secretion=inf, inplace=False):
        """
        Generate default environmental conditions for a given model
        Arguments:
            model (REFRAME or COBRAPY model or Simulator): model
            max_uptake (float): maximum uptake rate (default: 10.0)
            max_secretion (float): maximum secretion rate (default: 1000.0)
            inplace (bool): apply to model (default: False)
        Returns:
            Environment: Default environment for provided model
        """
        if not isinstance(model, Simulator):
            sim = get_simulator(model)
        else:
            sim = model

        env = Environment()

        for r_id in sim.get_exchange_reactions():
            env[r_id] = (-max_uptake, max_secretion)

        if inplace:
            env.apply(model, exclusive=False, inplace=True)
        else:
            return env

    @staticmethod
    def complete(model, max_uptake=10.0, inplace=False):
        """
        Generate a complete growth medium for a given model
        Arguments:
            model (REFRAME or COBRAPY model or Simulator): model
            max_uptake (float): maximum uptake rate (default: 1000.0)
            inplace (bool): apply to model (default: False)
        Returns:
            Environment: complete medium for provided model
        """

        return Environment.from_defaults(model, max_uptake=max_uptake, inplace=inplace)

    @staticmethod
    def empty(model, inplace=False):
        """
        Generate an empty growth medium for a given model
        Arguments:
            model (REFRAME or COBRAPY model or Simulator): model
            inplace (bool): apply to model (default: False)
        Returns:
            Environment: empty medium for provided model
        """

        return Environment.from_defaults(model, max_uptake=0, inplace=inplace)



from typing import TYPE_CHECKING, List, Dict, Iterable, Union

import pandas as pd

if TYPE_CHECKING:
    from .model_solution import ModelSolution


class MultiSolution:
    """
    A MultiSolution object is a collection of Solution objects.
    It can be used to compare different simulation methods or to compare the same method with different parameters.
    A MultiSolution object can be created by passing a list of Solution objects to the constructor.
    This object can be exported into a pandas DataFrame or Summary-like object
    using the to_frame(), to_summary() methods, respectively.
    """
    def __init__(self, *solutions: 'ModelSolution'):
        """
        A MultiSolution object is a collection of Solution objects.
        It can be used to compare different simulation methods or to compare the same method with different parameters.
        :param solutions: a list of Solution objects
        """
        if not solutions:
            _solutions = {}
        else:
            _solutions = {}

            for solution in solutions:
                setattr(self, f'{solution.method}', solution)
                _solutions[solution.method] = solution

        self._solutions = _solutions

    @property
    def solutions(self) -> Dict[str, 'ModelSolution']:
        """
        Returns a dict of Solution objects by the method name
        :return: a dict of Solution objects by the method name
        """
        return self._solutions

    def to_frame(self) -> pd.DataFrame:
        """
        Returns a pandas DataFrame with the results of the MultiSolution object
        :return: pandas DataFrame by the method name
        """
        frames = []
        columns = []

        for method, solution in self._solutions.items():
            frames.append(solution.to_frame().frame)
            columns.append(method)

        df = pd.concat(frames, axis=1, join='outer', keys=columns)

        return df

    def to_summary(self) -> pd.DataFrame:
        """
        Returns a pandas DataFrame with the summary of the MultiSolution object
        :return: pandas DataFrame by the method name
        """
        frames = []
        columns = []

        for method, solution in self._solutions.items():
            frames.append(solution.to_summary().frame)
            columns.append(method)

        df = pd.concat(frames, axis=1, join='outer', keys=columns)

        return df

    def __repr__(self):
        return 'MultiSolution'

    def __str__(self):
        return 'MultiSolution:' ','.join([method for method in self._solutions])


# TODO: methods stubs and type hinting
class DynamicSolution:
    """
    A DynamicSolution object is a collection of Solution objects.
    It is similar to the MultiSolution object, but it is used to store the results of a dynamic simulation using the
    time point rather than the method name.
    """
    def __init__(self, *solutions: 'ModelSolution', time: Iterable = None):
        """
        A DynamicSolution object is a collection of Solution objects.
        It is similar to the MultiSolution object, but it is used to store the results of a dynamic simulation using the
        time point rather than the method name.
        :param solutions: a list of Solution objects
        :param time: a linear space of time points
        """
        if not solutions:
            _solutions = {}
            time = []

        else:
            _solutions = {}

            if time is None:
                time = [i for i in range(len(solutions))]

            else:
                time = list(time)

            for t, solution in zip(time, solutions):
                setattr(self, f't_{t}', solution)
                _solutions[f't_{t}'] = solution

        self._solutions = _solutions
        self._time = time

    @property
    def solutions(self) -> Dict[str, 'ModelSolution']:
        """
        Returns a dict of Solution objects by the time point
        :return: a dict of Solution objects by the time point
        """
        return self._solutions

    def to_frame(self) -> pd.DataFrame:
        """
        Returns a pandas DataFrame with the results of the DynamicSolution object
        :return: pandas DataFrame by the time point
        """
        frames = []
        columns = []

        for time, solution in self._solutions.items():
            frames.append(solution.to_frame().frame)
            columns.append(time)

        df = pd.concat(frames, axis=1, join='outer', keys=columns)

        return df

    def to_summary(self) -> pd.DataFrame:
        """
        Returns a pandas DataFrame with the summary of the DynamicSolution object
        :return: pandas DataFrame by the time point
        """
        frames = []
        columns = []

        for time, solution in self._solutions.items():
            frames.append(solution.to_summary().frame)
            columns.append(time)

        df = pd.concat(frames, axis=1, join='outer', keys=columns)

        return df

    def __repr__(self):
        return 'DynamicSolution'

    def __str__(self):
        return 'DynamicSolution:' ','.join([time for time in self._solutions])


# TODO: methods stubs and type hinting
class KOSolution:
    """
    A KOSolution object is a collection of Solution objects.
    It is similar to the MultiSolution object, but it is used to store the results of a KO simulations.
    """
    def __init__(self,
                 solutions: Union[List['ModelSolution'], Dict[str, 'ModelSolution']],
                 kos: List[str] = None):
        """
        A KOSolution object is a collection of Solution objects.
        It is similar to the MultiSolution object, but it is used to store the results of a KO simulations.
        :param solutions: a list of Solution objects or a dict of Solution objects by the KO name
        :param kos: a list of KO names
        """
        if not solutions:
            _solutions = {}
            kos = []

        else:
            _solutions = {}

            if not kos:

                if isinstance(solutions, dict):

                    kos = list(solutions.keys())
                    solutions = list(solutions.values())

                else:

                    kos = [i for i, _ in enumerate(solutions)]

            else:

                if isinstance(solutions, dict):
                    solutions = list(solutions.values())

            for ko, solution in zip(kos, solutions):
                setattr(self, f'ko_{ko}', solution)
                _solutions[f'ko_{ko}'] = solution

        self._solutions = _solutions
        self._time = kos

    @property
    def solutions(self) -> Dict[str, 'ModelSolution']:
        """
        Returns a dict of Solution objects by the KO name
        :return: a dict of Solution objects by the KO name
        """
        return self._solutions

    def to_frame(self) -> pd.DataFrame:
        """
        Returns a pandas DataFrame with the results of the KOSolution object
        :return: pandas DataFrame by the KO name
        """
        frames = []
        columns = []

        for ko, solution in self._solutions.items():
            frames.append(solution.to_frame().frame)
            columns.append(ko)

        df = pd.concat(frames, axis=1, join='outer', keys=columns)

        return df

    def to_summary(self) -> pd.DataFrame:
        """
        Returns a pandas DataFrame with the summary of the KOSolution object
        :return: pandas DataFrame by the KO name
        """
        frames = []
        columns = []

        for ko, solution in self._solutions.items():
            frames.append(solution.to_summary().frame)
            columns.append(ko)

        df = pd.concat(frames, axis=1, join='outer', keys=columns)

        return df

    def __repr__(self):
        return 'KOSolution'

    def __str__(self):
        return 'KOSolution:' ','.join([time for time in self._solutions])



from typing import Union, List, Iterable

from mewpy.solvers.solver import VarType

from .linear_utils import Node


class VariableContainer:

    def __init__(self,
                 name: str,
                 sub_variables: List[str],
                 lbs: List[Union[int, float]],
                 ubs: List[Union[int, float]],
                 variables_type: List[VarType]):

        """

        Internal use only

        A container for variables, since multiple linear variables might have to be created out of a single variable
        during problem building. It is the main object for variable management in the linear problem object

        :param name: the name of variable. It is used as key/identifier of the variable in the linear problem
        :param sub_variables: the name of all sub-variables.
        If there is a single variable that does not have sub-variables, this attribute should be filled
        with the name of the variable only
        :param lbs: a list of the lower bounds of all sub-variables
        :param ubs: a list of the upper bounds of all sub-variables
        :param variables_type: a list of the variable types of all sub-variables
        """

        if not name:
            name = None

        if not sub_variables:
            sub_variables = []

        if not lbs:
            lbs = []

        if not ubs:
            ubs = []

        if not variables_type:
            variables_type = []

        self.name = name
        self.sub_variables = sub_variables
        self.lbs = lbs
        self.ubs = ubs
        self.variables_type = variables_type

    def __len__(self):
        return len(self.sub_variables)

    def __str__(self):
        return f'Variable {self.name}'

    def __eq__(self, other: 'VariableContainer'):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __iter__(self):

        self.__i = -1
        self.__n = len(self.sub_variables)

        return self

    def __next__(self):

        self.__i += 1

        if self.__i < self.__n:
            return self.sub_variables[self.__i]

        raise StopIteration

    def keys(self):

        return (var for var in self.sub_variables)

    def values(self):

        return ((lb, ub, var_type) for lb, ub, var_type in zip(self.lbs, self.ubs, self.variables_type))

    def items(self):

        return ((var, (lb, ub, var_type)) for var, lb, ub, var_type in zip(self.sub_variables,
                                                                           self.lbs,
                                                                           self.ubs,
                                                                           self.variables_type))

    def to_node(self):

        return Node(value=self.name, length=len(self.sub_variables))


class ConstraintContainer:

    def __init__(self,
                 name,
                 coefs,
                 lbs,
                 ubs):

        """

        Internal use only

        A container for constraints, since multiple constraints might have to be created out of a single constraint
        during problem building. It is the main object for constraint management in the linear problem object.

        For instance, the linearization of a single gpr can yield multiple rows/constraints to be added
        to the linear problem.
        Nevertheless, if one wants to replace/remove this gpr,
        a full mapping of the rows/constraints that must be replaced/removed can be found here
        and in the rows linked listed engine of the linear problem.

        :param name: the name of variable. It is used as key/identifier of the constraint in the linear problem
        :param coefs: a list of dictionaries having the coefficients for the constraint.
        That is, each dictionary in the list stands for a row in the linear problem. Dictionaries of coefficients
        must contain variable identifier value/coef pairs
        :param lbs: a list of the lower bounds of all coefficients
        :param ubs: a list of the upper bounds of all coefficients
        """

        if not name:
            name = None

        if not coefs:
            coefs = []

        if not lbs:
            lbs = []

        if not ubs:
            ubs = []

        self.name = name
        self.coefs = coefs
        self.lbs = lbs
        self.ubs = ubs

    def __len__(self):
        return len(self.coefs)

    def __str__(self):
        return f'Constraint {self.name}'

    def __eq__(self, other: 'ConstraintContainer'):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __iter__(self):

        self.__i = -1
        self.__n = len(self.coefs)

        return self

    def __next__(self):

        self.__i += 1

        if self.__i < self.__n:
            return self.coefs[self.__i]

        raise StopIteration

    def keys(self):

        return (i for i in range(len(self.coefs)))

    def values(self):

        return ((coef, lb, ub) for coef, lb, ub in zip(self.coefs, self.lbs, self.ubs))

    def items(self):

        return ((i, (coef, lb, ub)) for i, (coef, lb, ub) in enumerate(zip(self.coefs, self.lbs, self.ubs)))

    def to_node(self):

        return Node(value=self.name, length=len(self.coefs))


def concat_constraints(constraints: Iterable[ConstraintContainer], name: str = None):
    """
    Internal use only.

    Concatenates a list of constraints into a single constraint container.
    :param constraints: a list of constraint containers
    :param name: the name of the new constraint container
    :return: a new constraint container
    """
    coefs = []
    lbs = []
    ubs = []

    for cnt in constraints:
        coefs.extend(cnt.coefs)
        lbs.extend(cnt.lbs)
        ubs.extend(cnt.ubs)

    return ConstraintContainer(name=name, coefs=coefs, lbs=lbs, ubs=ubs)



import json
import os
from typing import Union, TYPE_CHECKING

from mewpy.germ.models import Model
from mewpy.io.dto import DataTransferObject

from .engine import Engine

if TYPE_CHECKING:
    from mewpy.germ.models import RegulatoryModel, Model, MetabolicModel


class JSON(Engine):
    def __init__(self, io, config, model=None):

        """
        Engine for JSON files
        """
        super().__init__(io, config, model)

    @property
    def model_type(self):
        return 'metabolic'

    @property
    def model(self):

        if self._model is None:
            return Model(identifier='model')

        return self._model

    def open(self,  mode='r'):

        self._dto = DataTransferObject()

        if mode == 'r':
            if isinstance(self.io, str):

                if not os.path.exists(self.io):
                    raise OSError(f'{self.io} is not a valid input. Provide the path or file handler')

                self._io = open(self.io)

        elif mode == 'w':
            pass

        else:
            raise ValueError(f'{mode} mode is not recognized. Try one of the following: r, w')

    def parse(self):

        if self.dto is None:
            raise OSError('Model is not open')

        try:

            self.dto.model = json.load(self.io)

        except BaseException as exc:

            self.close()
            self.clean()

            raise exc

        if 'types' not in self.dto.model:
            self.close()
            self.clean()

            raise OSError(f'{self.io} is not a valid json GERM model')

    def read(self,
             model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
             variables=None):

        return Model.from_dict(self.dto.model, variables=True)

    def write(self):

        dict_model = self.model.to_dict(variables=True)

        if hasattr(self.io, 'close'):

            try:
                json.dump(dict_model, self.io)

            except BaseException as exc:

                self.close()
                self.clean()

                raise exc

        else:
            with open(self.io, 'w') as file_path:
                json.dump(dict_model, file_path)

            self.clean()

    def close(self):

        if hasattr(self.io, 'close'):
            self.io.close()

    def clean(self):
        self._dto = None



"""
Algebra Constants

This module contains all algebra constants, namely operators and operand variables.
In addition, one can also find in this module the corresponding string representations of each operator and operand.
String representations will be used during expression parsing and evaluation, using python built-in eval in safe mode.

This module also contains common escape characters which will be replaced by the descriptive naming during expression
parsing.
"""

# Algebra operators and operands objects
from .symbolic import (BoolTrue,
                       BoolFalse,
                       And,
                       Or,
                       Not,
                       Equal,
                       NotEqual,
                       Inequality,
                       Greater,
                       GreaterEqual,
                       Less,
                       LessEqual,
                       Integer,
                       Float,
                       Zero,
                       One,
                       Symbol,
                       NoneAtom)

# Boolean algebra escape chars to be replaced
BOOLEAN_ESCAPE_CHARS = {'-': '_dash_',
                        ',': '_comma_',
                        ';': '_semicolon_',
                        '[': '_lbracket_',
                        ']': '_rbracket_',
                        ':': '_colon_',
                        '+': '_plus_',
                        '/': '_slash_',
                        '\\': '_backslash_',
                        '$': '_dollar_',
                        '%': '_percentage_',
                        '"': '_quotes_',
                        '(e)': '_e_',
                        '(c)': '_c_',
                        '(p)': '_p_'}

# Relational algebra escape chars to be replaced
RELATIONAL_ESCAPE_CHARS = BOOLEAN_ESCAPE_CHARS

# Main string representations for Boolean algebra operators
AND = '&'
OR = '|'
NOT = '~'

# Other string representations for Boolean algebra operators
BOOLEAN_OPERATORS = {'not': NOT,
                     '!': NOT,
                     'and': AND,
                     'or': OR,
                     AND: AND,
                     OR: OR,
                     NOT: NOT}

# Main string representations for Boolean algebra operators (set)
GLOBAL_BOOLEAN_OPERATORS = {AND, OR, NOT}

# Main string representations for Boolean algebra operands (except symbolic variables)
TRUE = '1'
FALSE = '0'

# Other string representations for Boolean algebra operands (except symbolic variables)
BOOLEAN_STATES = {'on': TRUE,
                  'off': FALSE,
                  'true': TRUE,
                  'false': FALSE,
                  TRUE: TRUE,
                  FALSE: FALSE}

# Main string representations for Boolean algebra operands (set)
GLOBAL_BOOLEAN_STATES = {TRUE, FALSE}

# Main string representations for Strict Relational algebra operators
GREATER = '>'
LESS = '<'
EQUAL = '='
NOT_EQUAL = '!='

# Other string representations for Strict Relational algebra operators
RELATIONAL_OPERATORS = {'greater than': GREATER,
                        'less than': LESS,
                        'greater': GREATER,
                        'less': LESS,
                        '==': EQUAL,
                        'equal': EQUAL,
                        'not equal': NOT_EQUAL,
                        GREATER: GREATER,
                        LESS: LESS,
                        EQUAL: EQUAL,
                        NOT_EQUAL: NOT_EQUAL}

# Main string representations for Strict Relational algebra operators (set)
GLOBAL_RELATIONAL_OPERATORS = {GREATER, LESS, EQUAL, NOT_EQUAL}

# Main string representations for Non-Strict Relational algebra operators
GREATER_EQUAL = '>='
LESS_EQUAL = '=>'

# Other string representations for Non-Strict Relational algebra operators
RELATIONAL_EQUAL_OPERATORS = {'greater than or equal': GREATER_EQUAL,
                              'less than or equal': LESS_EQUAL,
                              'greater or equal': GREATER_EQUAL,
                              'less or equal': LESS_EQUAL,
                              GREATER_EQUAL: GREATER_EQUAL,
                              LESS_EQUAL: LESS_EQUAL}

# Main string representations for Non-Strict Relational algebra operators (set)
GLOBAL_RELATIONAL_EQUAL_OPERATORS = {GREATER_EQUAL, LESS_EQUAL}

# Main string representations for Non-Strict and Strict algebra operands (except symbolic variables)
RELATIONAL_STATES = BOOLEAN_STATES

# Main string representations for all algebra operands (except symbolic variables)
GLOBAL_RELATIONAL_STATES = GLOBAL_BOOLEAN_STATES

# Main string representations for Non-Strict and Strict algebra operators and operands including empty, numeric and
# symbolic variables
GLOBAL_MEWPY_OPERATORS = {'BoolFalse': BoolFalse,
                          'BoolTrue': BoolTrue,
                          'And': And,
                          'Or': Or,
                          'Not': Not,
                          'Equal': Equal,
                          'NotEqual': NotEqual,
                          'Inequality': Inequality,
                          'Greater': Greater,
                          'GreaterEqual': GreaterEqual,
                          'Less': Less,
                          'LessEqual': LessEqual,
                          'Integer': Integer,
                          'Float': Float,
                          'Zero': Zero,
                          'One': One,
                          'Symbol': Symbol,
                          'NoneAtom': NoneAtom}



from mewpy.simulation import get_simulator
from mewpy.simulation.simulation import Simulator
from mewpy.util.parsing import build_tree, Boolean
from mewpy.util import AttrDict
from copy import deepcopy
from warnings import warn
from numpy import inf


class CommunityModel:
    """
    Community Model.

    :param models: A list of metabolic models.
    :param list fevaluation: A list of callable EvaluationFunctions.

    Optional parameters:
    :param bool copy_models: if the models are to be copied, default True.
    :param str flavor: use 'cobrapy' or 'reframed. Default 'cobrapy'. 

    """

    def __init__(self, models: list, copy_models=False, flavor='cobrapy'):
        self.organisms = AttrDict()
        self.model_ids = list({model.id for model in models})
        self.flavor = flavor
        self.biomasses = None
        self.biomass = None
        self.reaction_map = None
        self.metabolite_map = None

        if len(self.model_ids) < len(models):
            warn("Model ids are not unique, repeated models will be discarded.")

        for model in models:
            m = model if isinstance(model, Simulator) else get_simulator(model)
            self.organisms[m.id] = deepcopy(m) if copy_models else m

        sid = ' '.join(sorted(self.model_ids))
        if flavor == 'reframed':
            from reframed.core.cbmodel import CBModel
            model = CBModel(sid)
            self._met_prefix = 'M_'
            self._rxn_prefix = 'R_'
        else:
            from cobra.core.model import Model
            model = Model(sid)
            self._met_prefix = ''
            self._rxn_prefix = ''

        self.comm_model = get_simulator(model)
        self._merge_models()

    def get_community_model(self):
        return self.comm_model

    def size(self):
        return len(self.organisms)

    @property
    def merged_model(self):
        """ To make compatible with REFRAMED"""
        if self.comm_model is None:
            self._merged_model()
        return self.comm_model.model

    def _merge_models(self):
        """Merges the models. 
        """
        old_ext_comps = []
        ext_mets = []
        self.biomasses = {}
        self.reaction_map = {}
        self.metabolite_map = {}

        # default IDs
        ext_comp_id = "ext"
        biomass_id = "community_biomass"
        comm_growth = "community_growth"

        # create external compartment
        self.comm_model.add_compartment(ext_comp_id, "extracellular environment", external=True)

        # community biomass
        self.comm_model.add_metabolite(biomass_id, name="Total community biomass", compartment=ext_comp_id)

        self.comm_model.add_reaction(comm_growth, name="Community growth rate",
                                     stoichiometry={biomass_id: -1},
                                     lb=0, ub=inf)

        # add each organism

        for org_id, model in self.organisms.items():
            def rename(old_id):
                return f"{old_id}_{org_id}"

            # add internal compartments

            for c_id in model.compartments:
                comp = model.get_compartment(c_id)
                if comp.external:
                    old_ext_comps.append(c_id)
                else:
                    self.comm_model.add_compartment(rename(c_id), name=comp.name)

            # add metabolites

            for m_id in model.metabolites:
                met = model.get_metabolite(m_id)
                if met.compartment not in old_ext_comps:  # if is internal
                    new_id = rename(m_id)
                    self.comm_model.add_metabolite(new_id,
                                                   formula=met.formula,
                                                   name=met.name,
                                                   compartment=rename(met.compartment)
                                                   )
                    self.metabolite_map[(org_id,m_id)] = new_id

                elif m_id not in self.comm_model.metabolites:  # if is external but was not added yet
                    self.comm_model.add_metabolite(m_id,
                                                   formula=met.formula,
                                                   name=met.name,
                                                   compartment=ext_comp_id)
                    ext_mets.append(m_id)

            # add genes
            if self.flavor == 'reframed':
                for g_id in model.genes:
                    gene = model.get_gene(g_id)
                    new_id = rename(g_id)
                    self.comm_model.add_gene(new_id, gene.name)

            # add internal reactions
            ex_rxns = model.get_exchange_reactions()
            for r_id in model.reactions:
                rxn = model.get_reaction(r_id)
                if r_id in ex_rxns:
                    continue

                new_id = rename(r_id)
                new_stoichiometry = {
                    m_id if m_id in ext_mets else rename(m_id): coeff
                    for m_id, coeff in rxn.stoichiometry.items()
                }

                if r_id in [x for x, v in model.objective.items() if v > 0]:
                    new_stoichiometry[biomass_id] = 1
                    self.biomasses[org_id]=new_id

                if rxn.gpr:
                    t = build_tree(rxn.gpr, Boolean)
                    ren = {x:rename(x) for x in t.get_operands()}
                    new_gpr = t.replace(ren).to_infix().replace(' | ', ' or ').replace(' & ',' and ')
                else:
                    new_gpr = rxn.gpr
                self.comm_model.add_reaction(new_id,
                                             name=rxn.name,
                                             stoichiometry=new_stoichiometry,
                                             lb=rxn.lb,
                                             ub=rxn.ub,
                                             gpr=new_gpr)
                self.reaction_map[(org_id, r_id)] = new_id
        # Add exchange reactions

        for m_id in ext_mets:
            m = m_id[len(self._met_prefix):] if m_id.startswith(self._met_prefix) else m_id
            r_id = f"{self._rxn_prefix}EX_{m}"
            # self.comm_model.add_reaction(r_id, name=f'{m_id} exchange', stoichiometry={m_id: -1}, lb=-1000, ub=inf)
            self.comm_model.add_reaction(r_id, name=r_id, stoichiometry={m_id: -1}, lb=-inf, ub=inf)

        self.comm_model.objective = comm_growth
        self.biomass = comm_growth
        return self.comm_model



__author__ = 'BiSBII CEB University of Minho'
__email__ = 'vpereira@ceb.uminho.pt'
__version__ = '0.1.21'


def info():
    
    print('MEWpy version:',__version__)
    print('Author:',__author__)
    print('Contact:',__email__,'\n')

    from .simulation import __MEWPY_sim_solvers__, get_default_solver
    print('Available LP solvers:',' '.join(__MEWPY_sim_solvers__))
    print('Default LP solver:',get_default_solver(),'\n')
    
    from .solvers import __MEWPY_ode_solvers__, get_default_ode_solver
    print('Available ODE solvers:',' '.join(list(__MEWPY_ode_solvers__.keys())))
    print('Default ODE solver:', get_default_ode_solver(),'\n')
    
    from mewpy.util.utilities import get_all_subclasses
    from mewpy.problems.problem import AbstractProblem
    c = get_all_subclasses(AbstractProblem)
    print('Optimization Problems:',' '.join(sorted([x.__name__ for x in c])))
    print()
    from mewpy.optimization import engines, get_default_engine, get_available_algorithms
    print('Available EA engines:',' '.join(list(engines.keys())))
    print('Default EA engine:', get_default_engine())
    print('Available EAs:', ' '.join(sorted(get_available_algorithms())),'\n')
    


from typing import Union, Dict, TYPE_CHECKING, Any, Sequence, Tuple

import pandas as pd

from mewpy.germ.analysis import FBA
from mewpy.germ.solution import ModelSolution, KOSolution
from mewpy.solvers.solution import Solution, Status
from mewpy.solvers.solver import Solver
from mewpy.util.constants import ModelConstants

if TYPE_CHECKING:
    from mewpy.germ.variables import Regulator, Gene, Target
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


def _run_and_decode_solver(lp,
                           additional_constraints: Dict[str, Tuple[float, float]] = None,
                           **kwargs):
    if not additional_constraints:
        additional_constraints = {}

    if not kwargs:
        kwargs = {}

    if 'constraints' in kwargs:
        kwargs['constraints'].update(additional_constraints)

    solution = lp.solver.solve(**kwargs)
    if solution.status == Status.OPTIMAL:
        return solution.fobj
    else:
        return


class PROM(FBA):

    def __init__(self,
                 model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                 solver: Union[str, Solver] = None,
                 build: bool = False,
                 attach: bool = False):
        """
        The Probabilistic Regulation of Metabolism (PROM) algorithm predicts the growth phenotype and the flux response
        after transcriptional perturbation, given a metabolic and regulatory network.
        PROM introduces probabilities to represent gene states and gene-transcription factor interactions.

        For more detail consult: https://doi.org/10.1073/pnas.1005139107
        :param model: The metabolic and regulatory model to be simulated.
        :param solver: The solver to be used. If None, a new instance will be created from the default solver.
        :param build: If True, the linear problem will be built upon initialization.
        If False, the linear problem can be built later by calling the build() method.
        :param attach: If True, the linear problem will be attached to the model.
        """
        super().__init__(model=model, solver=solver, build=build, attach=attach)

    def _build(self):
        """
        It builds the PROM problem. It also builds a regular FBA problem to be used for the growth prediction.
        The linear problem is then loaded into the solver.
        :return:
        """
        self._build_mass_constraints()
        self._linear_objective = {var.id: value for var, value in self.model.objective.items()}
        self._minimize = False

    def _max_rates(self, solver_kwargs: Dict[str, Any]):
        # wt-type reference
        reference = self.solver.solve(**solver_kwargs)
        if reference.status != Status.OPTIMAL:
            raise RuntimeError('The solver did not find an optimal solution for the wild-type conditions.')
        reference = reference.values.copy()
        reference_constraints = {key: (reference[key] * 0.99, reference[key])
                                 for key in self._linear_objective}

        # fva of the reaction at fraction of 0.99 (for wild-type growth rate)
        rates = {}
        for reaction in self.model.reactions:
            min_rxn = _run_and_decode_solver(self,
                                             additional_constraints=reference_constraints,
                                             **{**solver_kwargs,
                                                'get_values': False,
                                                'linear': {reaction: 1},
                                                'minimize': True})
            max_rxn = _run_and_decode_solver(self,
                                             additional_constraints=reference_constraints,
                                             **{**solver_kwargs,
                                                'get_values': False,
                                                'linear': {reaction: 1},
                                                'minimize': False})

            reference_rate = reference[reaction]

            if reference_rate < 0:
                value = min((min_rxn, max_rxn, reference_rate))

            elif reference_rate > 0:
                value = max((min_rxn, max_rxn, reference_rate))

            else:
                value = max((abs(min_rxn), abs(max_rxn), abs(reference_rate)))

            if abs(value) < ModelConstants.TOLERANCE:
                value = 0.0

            rates[reaction] = value

        return rates

    def _optimize_ko(self,
                     probabilities: Dict[Tuple[str, str], float],
                     regulator: Union['Gene', 'Regulator'],
                     reference: Dict[str, float],
                     max_rates: Dict[str, float],
                     to_solver: bool = False,
                     solver_kwargs: Dict[str, Any] = None):
        solver_constrains = solver_kwargs.get('constraints', {})

        prom_constraints = {reaction.id: reaction.bounds for reaction in self.model.yield_reactions()}
        state = {gene: 1 for gene in self.model.genes.keys()}

        # if the regulator to be ko is a metabolic gene, the associated reactions are ko too
        # prom constraints of the associated reactions are set to threshold
        if regulator.is_gene():

            for reaction in regulator.reactions.keys():
                prom_constraints[reaction] = (-ModelConstants.TOLERANCE, ModelConstants.TOLERANCE)

        # finds the target genes of the deleted regulator.
        # finds the reactions associated with these target genes.
        # The reactions' bounds might be changed next, but for now the flag is set to False
        target_reactions = {}
        for target in regulator.yield_targets():

            if target.is_gene():
                # after the regulator ko iteration, this is reset
                state[target.id] = 0

                target_reactions.update({reaction.id: reaction for reaction in target.yield_reactions()})

        # GPR evaluation of each reaction previously found, but using the changed gene_state.
        # If the GPR is evaluated to zero, the reaction bounds will be changed in the future.
        # For that, the reactions dictionary flags must be updated to True.
        inactive_reactions = {}
        for reaction in target_reactions.values():

            if reaction.gpr.is_none:
                continue

            if reaction.gpr.evaluate(values=state):
                continue

            inactive_reactions[reaction.id] = reaction

        # for each target regulated by the regulator
        for target in regulator.yield_targets():

            if not target.is_gene():
                continue

            target: Union['Target', 'Gene']

            # composed key for interactions_probabilities
            target_regulator = (target.id, regulator.id)

            if target_regulator not in probabilities:
                continue

            interaction_probability = probabilities[target_regulator]

            # for each reaction associated with this single target
            for reaction in target.yield_reactions():

                # if the gpr has been evaluated previously to zero,
                # it means that the metabolic genes regulated by this regulator can affect the state of the
                # reaction. Thus, the reaction bounds can be changed using PROM probability.
                # Nevertheless, it is only useful to do that if the probability is inferior to 1, otherwise
                # nothing is changed
                if reaction.id not in inactive_reactions:
                    continue

                if interaction_probability >= 1:
                    continue

                # reaction old bounds
                rxn_lb, rxn_ub = tuple(prom_constraints[reaction.id])

                # probability flux is the upper or lower bound that this reaction can take
                # when the regulator is KO. This is calculated as follows:
                # interaction probability times the reaction maximum limit (determined by fva)
                probability_flux = max_rates[reaction.id] * interaction_probability

                # wild-type flux value for this reaction
                wt_flux = reference[reaction.id]

                # update flux bounds according to probability flux
                if wt_flux < 0:

                    rxn_lb = max((reaction.lower_bound, probability_flux, rxn_lb))
                    rxn_lb = min((rxn_lb, -ModelConstants.TOLERANCE))

                elif wt_flux > 0:

                    rxn_ub = min((reaction.upper_bound, probability_flux, rxn_ub))
                    rxn_ub = max((rxn_ub, ModelConstants.TOLERANCE))

                else:

                    # if it is zero, the reaction is not changed, so that reactions are not activated
                    # by PROM. Only reaction ko is forced by PROM.

                    continue

                prom_constraints[reaction.id] = (rxn_lb, rxn_ub)

        solution = self.solver.solve(**{**solver_kwargs,
                                        'get_values': True,
                                        'constraints': {**solver_constrains, **prom_constraints}})

        if to_solver:
            return solution

        minimize = solver_kwargs.get('minimize', self._minimize)
        return ModelSolution.from_solver(method=self.method, solution=solution, model=self.model,
                                         minimize=minimize)

    def _optimize(self,
                  initial_state: Dict[Tuple[str, str], float] = None,
                  regulators: Sequence[Union['Gene', 'Regulator']] = None,
                  to_solver: bool = False,
                  solver_kwargs: Dict[str, Any] = None) -> Union[Dict[str, Solution], Dict[str, ModelSolution]]:
        # wild-type reference
        solver_kwargs['get_values'] = True
        reference = self.solver.solve(**solver_kwargs)
        if reference.status != Status.OPTIMAL:
            raise RuntimeError('The solver did not find an optimal solution for the wild-type conditions.')
        reference = reference.values.copy()

        # max and min fluxes of the reactions
        max_rates = self._max_rates(solver_kwargs=solver_kwargs)

        # a single regulator knockout
        if len(regulators) == 1:
            return self._optimize_ko(probabilities=initial_state,
                                     regulator=regulators[0],
                                     reference=reference,
                                     max_rates=max_rates,
                                     to_solver=to_solver,
                                     solver_kwargs=solver_kwargs)

        # multiple regulator knockouts
        kos = {}
        for regulator in regulators:
            ko_solution = self._optimize_ko(probabilities=initial_state,
                                            regulator=regulator,
                                            reference=reference,
                                            max_rates=max_rates,
                                            to_solver=to_solver,
                                            solver_kwargs=solver_kwargs)
            kos[regulator.id] = ko_solution
        return kos

    def optimize(self,
                 initial_state: Dict[Tuple[str, str], float] = None,
                 regulators: Union[str, Sequence['str']] = None,
                 to_solver: bool = False,
                 solver_kwargs: Dict[str, Any] = None) -> Union[KOSolution, Dict[str, Solution]]:
        """
        It solves the PROM linear problem. The linear problem is solved using the solver interface.

        The optimize method allows setting temporary changes to the linear problem. The changes are
        applied to the linear problem reverted to the original state afterward.
        Objective, constraints and solver parameters can be set temporarily.
        :param initial_state: dictionary with the probabilities of
        the interactions between the regulators and the targets.
        :param regulators: list of regulators to be knocked out. If None, all regulators are knocked out.
        :param to_solver: Whether to return the solution as a SolverSolution instance. Default: False.
        Otherwise, a ModelSolution is returned.
        :param solver_kwargs: Solver parameters to be set temporarily.
            - linear: A dictionary of linear coefficients to be set temporarily. The keys are the variable names
            and the values are the coefficients. Default: None
            - quadratic: A dictionary of quadratic coefficients to be set temporarily. The keys are tuples of
            variable names and the values are the coefficients. Default: None
            - minimize: Whether to minimize the objective. Default: False
            - constraints: A dictionary with the constraints bounds. The keys are the constraint ids and the values
            are tuples with the lower and upper bounds. Default: None
            - get_values: Whether to retrieve the solution values. Default: True
            - shadow_prices: Whether to retrieve the shadow prices. Default: False
            - reduced_costs: Whether to retrieve the reduced costs. Default: False
            - pool_size: The size of the solution pool. Default: 0
            - pool_gap: The gap between the best solution and the worst solution in the pool. Default: None
        :return: A KOSolution instance or a list of SolverSolution instance if to_solver is True.
        """
        # build solver if out of sync
        if not self.synchronized:
            self.build()

        if not initial_state:
            initial_state = {}

        if not regulators:
            regulators = list(self.model.yield_regulators())
        else:
            if isinstance(regulators, str):
                regulators = [regulators]

            regulators = [self.model.get(regulator) for regulator in regulators]

        if not solver_kwargs:
            solver_kwargs = {}

        # concrete optimize
        solutions = self._optimize(initial_state=initial_state,
                                   regulators=regulators,
                                   to_solver=to_solver,
                                   solver_kwargs=solver_kwargs)

        if to_solver:
            return solutions

        return KOSolution(solutions)


# ----------------------------------------------------------------------------------------------------------------------
# Probability of Target-Regulator interactions
# ----------------------------------------------------------------------------------------------------------------------
def target_regulator_interaction_probability(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                                             expression: pd.DataFrame,
                                             binary_expression: pd.DataFrame) -> Tuple[Dict[Tuple[str, str], float],
                                                                                       Dict[Tuple[str, str], float]]:
    """
    It computes the conditional probability of a target gene being active when the regulator is inactive.
    It uses the following formula:
        P(target = 1 | regulator = 0) = count(target = 1, regulator = 0) / # samples
    This probability is computed for each combination of target-regulator.
    This method is used in PROM analysis.

    :param model: an integrated Metabolic-Regulatory model aka a GERM model
    :param expression: Quantile preprocessed expression matrix
    :param binary_expression: Quantile preprocessed expression matrix binarized
    :return: Dictionary with the conditional probability of a target gene being active when the regulator is inactive,
    Dictionary with missed interactions
    """
    try:
        # noinspection PyPackageRequirements
        from scipy.stats import ks_2samp
    except ImportError:
        raise ImportError('The package scipy is not installed. '
                          'To compute the probability of target-regulator interactions, please install scipy '
                          '(pip install scipy).')
    missed_interactions = {}
    interactions_probabilities = {}

    for interaction in model.yield_interactions():

        target = interaction.target

        if not interaction.regulators or target.id not in expression.index:
            missed_interactions[(target.id, target.id)] = 1
            interactions_probabilities[(target.id, target.id)] = 1
            continue

        target_expression = expression.loc[target.id]
        target_binary = binary_expression.loc[target.id]

        for regulator in interaction.yield_regulators():

            if regulator.id not in expression.index:
                missed_interactions[(target.id, regulator.id)] = 1
                interactions_probabilities[(target.id, regulator.id)] = 1
                continue

            regulator_binary = binary_expression.loc[regulator.id]

            target_expression_1_regulator = target_expression[regulator_binary == 1]
            target_expression_0_regulator = target_expression[regulator_binary == 0]

            if len(target_expression_1_regulator) == 0 and len(target_expression_0_regulator) == 0:
                missed_interactions[(target.id, regulator.id)] = 1
                interactions_probabilities[(target.id, regulator.id)] = 1
                continue

            _, p_val = ks_2samp(target_expression_1_regulator, target_expression_0_regulator)
            if p_val < 0.05:
                target_binary_0_regulator = target_binary[regulator_binary == 0]

                probability = sum(target_binary_0_regulator) / len(target_binary_0_regulator)

                interactions_probabilities[(target.id, regulator.id)] = probability
                missed_interactions[(target.id, regulator.id)] = 0

            else:
                missed_interactions[(target.id, regulator.id)] = 1
                interactions_probabilities[(target.id, regulator.id)] = 1

    return interactions_probabilities, missed_interactions



from symbol import parameters
from mewpy.util.parsing import Arithmetic, build_tree
from mewpy.util.utilities import AttrDict
from collections import OrderedDict
import warnings
import numpy as np

class Compartment(object):
    """ class for modeling compartments. """

    def __init__(self, comp_id, name=None, external=False, size=1.0):
        """
        Arguments:
            comp_id (str): a valid unique identifier
            name (str): compartment name (optional)
            external (bool): is external (default: false)
            size (float): compartment size (default: 1.0)
        """
        self.id = comp_id
        self.name = name if name is not None else comp_id
        self.size = size
        self.external = external
        self.metadata = OrderedDict()

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)


class Metabolite(object):
    """ class for modeling metabolites. """

    def __init__(self, met_id, name=None, compartment=None):
        """
        Arguments:
            met_id (str): a valid unique identifier
            name (str): common metabolite name
            compartment (str): compartment containing the metabolite
        """
        self.id = met_id
        self.name = name if name is not None else met_id
        self.compartment = compartment
        self.metadata = OrderedDict()

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)


def calculate_yprime(y, rate, substrates, products, substrate_names):
    """
    It takes the numpy array for y_prime,
    and adds or subtracts the amount in rate to all the substrates or products listed
    Returns the new y_prime
    Args:
        y: a numpy array for the substrate values, the same order as y
        rate: the rate calculated by the user made rate equation
        substrates: list of substrates for which rate should be subtracted
        products: list of products for which rate should be added
        substrate_names: the ordered list of substrate names in the model.\
            Used to get the position of each substrate or product in y_prime
    Returns:
        y_prime: following the addition or subtraction of rate to the specificed substrates
    """
    y_prime = np.zeros(len(y))

    for name in substrates:
        y_prime[name] -= rate

    for name in products:
        y_prime[name] += rate

    return y_prime


def check_positive(y_prime):
    """
    Check that substrate values are not negative when they shouldnt be.
    """

    for i in range(len(y_prime)):
        if y_prime[i] < 0:
            y_prime[i] = 0

    return y_prime


class Rule(object):
    """Base class for kinetic rules.
    """

    def __init__(self, r_id, law: str, parameters: dict = {}):
        """Creates a new rule

        Args:
            r_id (str): Reaction/rule identifier
            law (str): The rule string representation.
        """
        self.id = r_id
        self.law = law
        self._tree = None
        self.parameters = parameters

    @property
    def tree(self):
        """Parsing tree of the law.

        Returns:
            Node: Root node of the parsing tree.
        """
        if not self._tree:
            self._tree = build_tree(self.law, Arithmetic)
        return self._tree        

    def parse_parameters(self):
        """Returns the list of parameters within the rule.

        Returns:
            list: parameters
        """
        return list(self.tree.get_parameters())

    def rename_parameter(self, old_parameter, new_parameter):
        """Need to rename in the law, tree and parameters"""

        if old_parameter in self.parse_parameters():
            t = self.tree.replace({old_parameter:new_parameter})
            self.law = t.to_infix()
            self._tree = None
        try:
            self.parameters[new_parameter]=self.parameters[old_parameter]
            del self.parameters[old_parameter]
        except:
            pass



    def get_parameters(self):
        return self.parameters

    def replace(self, parameters=None, local=True, infix=True):
        """Replaces parameters with values taken from a dictionary.
        If no parameter are given for replacement, returns the string representation of the rule
        built from the parsing tree.

        Args:
            parameters (dict, optional): Replacement dictionary. Defaults to None.
            local (bool, optional): use parameter values defined in the rule 
        Returns:
            str: the kinetic rule.
        """
        param = parameters.copy() if parameters else dict()
        if local:
            param.update(self.parameters)
        t = self.tree.replace(param)
        if infix:
            return t.to_infix()
        else:
            return t

    def calculate_rate(self, substrates={}, parameters={}):
        param = dict()
        param.update(self.parameters)
        param.update(substrates)
        param.update(parameters)

        if len(param.keys()) != len(self.parse_parameters()):
            s = set(self.parse_parameters())-set(param.keys())
            raise ValueError(f"Values missing for parameters: {s}")
        t = self.replace(param)
        rate = eval(t)
        return rate

    def __str__(self):
        return self.law.replace(' ','')

    def __repr__(self):
        return self.replace().replace(' ','')


class KineticReaction(Rule):

    def __init__(self, 
                 r_id:str,
                 law: str,
                 name:str=None, 
                 stoichiometry: dict = {}, 
                 parameters: dict = {}, 
                 modifiers: list = [],
                 reversible:bool=True):
        """Kinetic reaction rule.

        Args:
            r_id (str): Reaction identifier
            law (str): kinetic law
            stoichiometry (dict): The stoichiometry of the reaction.
            parameters (dict, optional): local parameters. Defaults to dict().
            substrates (list, optional): substrates. Defaults to [].
            products (list, optional): products. Defaults to [].
        """
        super(KineticReaction, self).__init__(r_id, law, parameters)
        self.name = name if name else r_id
        self.stoichiometry = stoichiometry
        self.modifiers = modifiers
        self.parameter_distributions = {}
        self.reversible = reversible
        self._model = None

    @property
    def substrates(self):
        return [k for k, v in self.stoichiometry.items() if v < 0]

    @property
    def products(self):
        return [k for k, v in self.stoichiometry.items() if v > 0]

    def rename_parameter(self, old_parameter, new_parameter):
        super().rename_parameter(old_parameter, new_parameter)

        if old_parameter in self.modifiers:
            self.modifiers.append(new_parameter)
            self.modifiers.remove(old_parameter)

        if old_parameter in self.parameter_distributions:
            self.parameter_distributions[new_parameter] = self.parameter_distributions[old_parameter]
            del self.parameter_distributions[old_parameter]
        

    def set_parameter_distribution(self, param, dist):
        self.parameter_distributions[param]=dist

    def sample_parameter(self,param):
        dist = self.parameter_distributions.get(param,None)
        if not dist:
            raise ValueError(f"The parameter {param} has no associated distribution.")
        return dist.rvs()

    def parse_law(self, map: dict, local=True):
        """Auxiliary method invoked by the model to build the ODE system.

        Args:
            map (dict): Dictionary of global paramameters replacements.

        Returns:
            str: kinetic rule.
        """
        m = {p_id: f"p['{p_id}']" for p_id in self.parameters.keys()}
        r_map = map.copy()
        r_map.update(m)

        return self.replace(r_map, local=local)

    def calculate_rate(self, substrates={}, parameters={}):
        
        param = dict()
        # sets model defaults 
        param.update(self._model.get_concentrations())
        param.update(self._model.get_parameters())
        # set reaction defaults
        param.update(self.parameters)
        # user defined
        param.update(substrates)
        param.update(parameters)
        s = set(self.parse_parameters())-set(param.keys())
        if s:
            # check for missing parameters distributions  
            r = s - set(self.parameter_distributions.keys())
            if r:
                raise ValueError(f"Missing values or distribuitions for parameters: {r}")
            else:
                for p in s:
                    param[p] = self.parameter_distributions[p].rvs()
        t = self.replace(param)
        rate = eval(t)
        return rate

    def reaction(self, y, substrate_names, parameter_dict):

        for modifier in self.modifiers:
            pass

        parameters = self.get_parameters(parameter_dict)
        substrates =self.substrates

        if len(self.modifiers) != 0:
            # substrates, parameters = self.calculate_modifiers(substrates, parameters)
            pass

        rate = self.calculate_rate(substrates, parameters)

        y_prime = calculate_yprime(y, rate, self.substrates, self.products, substrate_names)
        # y_prime = self.modify_product(y_prime, substrate_names)

        if not self.reversible:
            y_prime = check_positive(y_prime)

        return y_prime

    def set_parameter_defaults_to_mean(self):
        """Sets not defined parameters to the median of a distribution.
        """ 
        for name in self.parameter_distributions:
            if name not in self.parameters:
                if (type(self.parameter_distributions[name]) == list or
                        type(self.parameter_distributions[name]) == tuple):
                    self.parameters[name] = (self.parameter_distributions[name][0] +
                                             self.parameter_distributions[name][1]) / 2
                else:
                    self.parameters[name] = self.parameter_distributions[name].mean()


class ODEModel:
    def __init__(self, model_id):
        """ ODE Model.
        """
        self.id = model_id
        self.metabolites = OrderedDict()
        self.compartments = OrderedDict()
        # kinetic rule of each reaction
        self.ratelaws = OrderedDict()
        # initial concentration of metabolites
        self.concentrations = OrderedDict()
        # parameter defined as constantes
        self.constant_params = OrderedDict()
        # variable parameters
        self.variable_params = OrderedDict()
        self.assignment_rules = OrderedDict()

        self._func_str = None
        self._constants = None
        self._m_r_lookup = None

    def _clear_temp(self):
        self._func_str = None

    def add_compartment(self, compartment, replace=True):
        """ Add a compartment to the model.
        Arguments:
            compartment (Compartment): compartment to add
            replace (bool): replace previous compartment with same id (default: True)
        """
        if compartment.id in self.compartments and not replace:
            raise RuntimeError(f"Compartment {compartment.id} already exists.")
        self.compartments[compartment.id] = compartment

    def add_metabolite(self, metabolite, replace=True):
        """ Add a metabolite to the model.
        Arguments:
            metabolite (Metabolite): metabolite to add
            replace (bool): replace previous metabolite with same id (default: True)
        """

        if metabolite.id in self.metabolites and not replace:
            raise RuntimeError(f"Metabolite {metabolite.id} already exists.")

        if metabolite.compartment not in self.compartments:
            raise RuntimeError(f"Metabolite {metabolite.id} has invalid compartment {metabolite.compartment}.")

        self.metabolites[metabolite.id] = metabolite

    @property
    def reactions(self):
        return AttrDict(self.ratelaws)

    def get_reaction(self,r_id):
        if r_id not in self.ratelaws:
            raise ValueError(f'Unknown reaction {r_id}')
        r = self.ratelaws[r_id]
        d = {'id':r_id,
             'name':r.name,
             'stoichiometry':r.stoichiometry,
             'law':r.law,
             'reversible':r.reversible,
             'parameters':r.parameters,
             'modifiers':r.modifiers
             } 
        return AttrDict(d)

    def get_metabolite(self,m_id):
        if m_id not in self.metabolites:
            raise ValueError(f'Unknown metabolite {m_id}')
        d = { 'id':m_id,
              'name': self.metabolites[m_id].name,
              'compartment':self.metabolites[m_id].compartment,
              'formula':self.metabolites[m_id].metadata.get('FORMULA',''),
              'charge':self.metabolites[m_id].metadata.get('CHARGE',''),
              'y0': self.concentrations[m_id]
            }
        return AttrDict(d)


    def find(self, pattern=None, sort=False):
        """A user friendly method to find reactionsin the model.

        :param pattern: The pattern which can be a regular expression, defaults to None in which case all entries are listed.
        :type pattern: str, optional
        :param sort: if the search results should be sorted, defaults to False
        :type sort: bool, optional
        :return: the search results
        :rtype: pandas dataframe
        """
        values = list(self.reactions.keys())
        if pattern:
            import re
            if isinstance(pattern, list):
                patt = '|'.join(pattern)
                re_expr = re.compile(patt)
            else:
                re_expr = re.compile(pattern)
            values = [x for x in values if re_expr.search(x) is not None]
        if sort:
            values.sort()

        import pandas as pd
        data = [self.get_reaction(x) for x in values]
        
        if data:
            df = pd.DataFrame(data)
            df = df.set_index(df.columns[0])
        else: 
            df = pd.DataFrame()
        return df

    def find_reactions(self, pattern=None, sort=False):
        return self.find(pattern,sort)
    
    def find_metabolites(self, pattern=None, sort=False):
        """A user friendly method to find metabolites in the model.

        :param pattern: The pattern which can be a regular expression, defaults to None in which case all entries are listed.
        :type pattern: str, optional
        :param sort: if the search results should be sorted, defaults to False
        :type sort: bool, optional
        :return: the search results
        :rtype: pandas dataframe
        """
        values = list(self.metabolites.keys())
        if pattern:
            import re
            if isinstance(pattern, list):
                patt = '|'.join(pattern)
                re_expr = re.compile(patt)
            else:
                re_expr = re.compile(pattern)
            values = [x for x in values if re_expr.search(x) is not None]
        if sort:
            values.sort()

        import pandas as pd
        data = [self.get_metabolite(x) for x in values]
        
        if data:
            df = pd.DataFrame(data)
            df = df.set_index(df.columns[0])
        else: 
            df = pd.DataFrame()
        return df

    def get_concentrations(self):
        return self.concentrations

    def set_concentration(self, m_id: str, concentration: float):
        """Sets a metabolite initial concentration

        Args:
            m_id (str): Metabolite identifier
            concentration (float): Initial concentration
        """
        if m_id in self.metabolites:
            self.concentrations[m_id] = concentration
        else:
            warnings.warn(f"No such metabolite '{m_id}'", RuntimeWarning)

    def set_ratelaw(self, r_id: str, law: KineticReaction):
        """Define the rate law for a given reaction.

        Args:
            r_id (str): Reaction Identifier
            law (KineticReaction): The reaction rate law.
        """
        law._model = self
        self.ratelaws[r_id] = law

    def get_ratelaw(self, r_id):
        if r_id in self.ratelaws.keys():
            return self.ratelaws[r_id]
        else:
            raise ValueError('Reaction has no rate law.')

    def set_assignment_rule(self, p_id: str, rule: Rule):
        if p_id in self.variable_params or p_id in self.metabolites:
            self.assignment_rules[p_id] = rule
        else:
            warnings.warn(f"No such variable parameter '{p_id}'", RuntimeWarning)

    def set_global_parameter(self, key, value, constant=True):
        if constant:
            self.constant_params[key] = value
        else:
            self.variable_params[key] = value

    def merge_constants(self):
        constants = OrderedDict()

        for c_id, comp in self.compartments.items():
            constants[c_id] = comp.size

        constants.update(self.constant_params)

        for r_id, law in self.ratelaws.items():
            for p_id, value in law.parameters.items():
                
                full_id = f"{p_id}"
                if full_id in constants:
                    warnings.warn(f'renaming {p_id} to {r_id}_{p_id}')
                    full_id = f"{r_id}_{p_id}"
                    law.rename_parameter(f"{p_id}",f"{r_id}_{p_id}")
                constants[full_id] = value

        self._constants = constants
        return constants

    def metabolite_reaction_lookup(self):
        if not self._m_r_lookup:
            self._m_r_lookup = {m_id: {} for m_id in self.metabolites}

            for r_id, rule in self.ratelaws.items():
                for m_id, coeff in rule.stoichiometry.items():
                    self._m_r_lookup[m_id][r_id] = coeff
        return self._m_r_lookup

    def print_balance(self, m_id, factors=None):
        f = factors.get(m_id, 1) if factors else 1
        c_id = self.metabolites[m_id].compartment
        table = self.metabolite_reaction_lookup()

        terms = []
        for r_id, coeff in table[m_id].items():
            v = coeff * f if coeff > 0 else coeff
            terms.append(f"{v:+g} * r['{r_id}']")

        if ( f == 0 or 
             len(terms) == 0 or 
             ( self.metabolites[m_id].constant and 
               self.metabolites[m_id].boundary
             )):
            expr = "0"
        else:
            expr = f"1/p['{c_id}'] * ({' '.join(terms)})"
        return expr

    def get_parameters(self, exclude_compartments=False):
        """Returns a dictionary of the model parameters

        """
        if not self._constants:
            self.merge_constants()
        parameters = self._constants.copy()
        if exclude_compartments:
            for c_id in self.compartments:
                del parameters[c_id]
        return parameters

    def find_parameters(self, pattern=None, sort=False):
        """A user friendly method to find parameters in the model.

        :param pattern: The pattern which can be a regular expression, defaults to None in which case all entries are listed.
        :type pattern: str, optional
        :param sort: if the search results should be sorted, defaults to False
        :type sort: bool, optional
        :return: the search results
        :rtype: pandas dataframe
        """
        params = self.get_parameters()
        values = list(params.keys())
        if pattern:
            import re
            if isinstance(pattern, list):
                patt = '|'.join(pattern)
                re_expr = re.compile(patt)
            else:
                re_expr = re.compile(pattern)
            values = [x for x in values if re_expr.search(x) is not None]
        if sort:
            values.sort()

        import pandas as pd
        data = [(x,params[x]) for x in values]
        
        if data:
            df = pd.DataFrame(data,columns=['Parameter','Value'])
            df = df.set_index(df.columns[0])
        else: 
            df = pd.DataFrame()
        return df

    def deriv(self, y, t):
        """
        deriv function called by integrate

        For each step when the model is run, the rate for each reaction is calculated and changes in substrates and products calculated.
        These are returned by this function as y_prime, which are added to y which is returned by run_model

        Args:
            y (list): ordered list of substrate values at this current timepoint. Has the same order as self.run_model_species_names
            t (): time, not used in this function but required for some reason

        Returns:
            y_prime - ordered list the same as y, y_prime is the new set of y's for this timepoint.
        """
        m_y = OrderedDict(zip(self.metabolites, y))
        yprime = np.zeros(len(y))
        for _, reaction in self.ratelaws.items():
            yprime += reaction.reaction(m_y, self.get_parameters())
        return yprime

    def build_ode(self, factors:dict=None, local:bool=False) -> str:
        """ 
        Axiliary function to build the ODE as a string
        to be evaluated by eval. Allows the inclusion of factors to
        be applied to the parameters

        Args:
            factors (dict): factors to be applied to parameters
            local (bool): enforces the usage of parameter values defined within
            the reactions
        Returns:
            func_str the right-hand side of the system. 
        """
        rmap = OrderedDict()
        m = {m_id: f"x[{i}]" for i, m_id in enumerate(self.metabolites)}
        c = {c_id: f"p['{c_id}']" for c_id in self.compartments}
        p = {p_id: f"p['{p_id}']" for p_id in self.constant_params}
        v = {p_id: f"v['{p_id}']" for p_id in self.variable_params}
        rmap.update(m)
        rmap.update(c)
        rmap.update(p)
        rmap.update(v)
        
        parsed_rates = {r_id: ratelaw.parse_law(rmap, local=local)
                        for r_id, ratelaw in self.ratelaws.items()}

        r = {r_id: f"({parsed_rates[r_id]})" for r_id in self.ratelaws.keys()}

        rmap.update(r)

        rate_exprs = [' '*4+"r['{}'] = {}".format(r_id, parsed_rates[r_id])
                      for r_id in self.ratelaws.keys()]

        balances = [' '*8 + self.print_balance(m_id, factors=factors) for m_id in self.metabolites]

        func_str = 'def ode_func(t, x, r, p, v):\n\n' + \
            '\n'.join(rate_exprs) + '\n\n' + \
            '    dxdt = [\n' + \
            ',\n'.join(balances) + '\n' + \
            '    ]\n\n' + \
            '    return dxdt\n'

        self._func_str = func_str
        return self._func_str

    def get_ode(self, r_dict=None, params=None, factors=None):
        """
        Args:
            r_dict: reaction identifiers to be modified
            params: modified parameters
            factors: factors to be applied to parameters
        """
        p = self.merge_constants()
        if params:
            p.update(params)

        if factors is not None:
            for k,v in factors.items():
                if k in p.keys():
                    p[k]= v * p[k] 

        v = self.variable_params.copy()
        r = r_dict if r_dict is not None else dict()
        
        np.seterr(divide='ignore', invalid='ignore')         
        exec(self.build_ode(factors), globals())
        ode_func = eval('ode_func')

        return lambda t, x: ode_func(t, x, r, p, v)



from typing import Union, Dict

from mewpy.germ.lp import ConstraintContainer, VariableContainer, LinearProblem
from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel
from mewpy.solvers.solution import Solution
from mewpy.solvers.solver import VarType, Solver


class FBA(LinearProblem):

    def __init__(self,
                 model: Union[Model, MetabolicModel, RegulatoryModel],
                 solver: Union[str, Solver, None] = None,
                 build: bool = False,
                 attach: bool = False):
        """
        Flux Balance Analysis (FBA) of a metabolic model. Regular implementation of a FBA for a metabolic model.

        For more details consult: https://dx.doi.org/10.1038%2Fnbt.1614

        :param model: a MetabolicModel, RegulatoryModel or GERM model. The model is used to retrieve
        variables and constraints to the linear problem
        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used,
        but it will be overwritten if build is true.
        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """
        super().__init__(model=model, solver=solver, build=build, attach=attach)

    def _build_mass_constraints(self):
        gene_state = {gene.id: max(gene.coefficients) for gene in self.model.yield_genes()}

        constraints = {metabolite.id: ConstraintContainer(name=metabolite.id, lbs=[0.0], ubs=[0.0], coefs=[{}])
                       for metabolite in self.model.yield_metabolites()}
        variables = {}

        for reaction in self.model.yield_reactions():
            if reaction.gpr.is_none:
                lb, ub = reaction.bounds

            else:
                res = reaction.gpr.evaluate(values=gene_state)
                if not res:
                    lb, ub = 0.0, 0.0
                else:
                    lb, ub = reaction.bounds

            variable = VariableContainer(name=reaction.id, sub_variables=[reaction.id],
                                         lbs=[float(lb)], ubs=[float(ub)], variables_type=[VarType.CONTINUOUS])
            variables[reaction.id] = variable

            for metabolite, stoichiometry in reaction.stoichiometry.items():
                constraints[metabolite.id].coefs[0][reaction.id] = stoichiometry

        self.add_variables(*variables.values())
        self.add_constraints(*constraints.values())
        return

    def _build(self):
        """
        It builds the linear problem from the model. The linear problem is built from the model
        variables and constraints. The linear problem is then loaded into the solver.
        :return:
        """
        if self.model.is_metabolic():
            # mass balance constraints and reactions' variables
            self._build_mass_constraints()

            self._linear_objective = {var.id: value for var, value in self.model.objective.items()}
            self._minimize = False

        return

    def _optimize(self, solver_kwargs: Dict = None, **kwargs) -> Solution:
        """
        It optimizes the linear problem. The linear problem is solved by the solver interface.
        :param solver_kwargs: A dictionary of keyword arguments to be passed to the solver.
        :return: A Solution instance.
        """
        return self.solver.solve(**solver_kwargs)



from typing import Any, Dict, Union, TYPE_CHECKING, Tuple, Generator, List

from mewpy.germ.algebra import Expression, parse_expression
from mewpy.util.utilities import generator
from mewpy.germ.models.serialization import serialize
from mewpy.util.history import recorder
from mewpy.util.constants import ModelConstants
from mewpy.io.engines.engines_utils import expression_warning
from .variable import Variable, variables_from_symbolic

if TYPE_CHECKING:
    from .metabolite import Metabolite
    from .gene import Gene
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


def _bounds_setter(instance, old_bounds):
    instance._bounds = tuple(old_bounds)

    if instance.model:
        instance.model.notify()


class Reaction(Variable, variable_type='reaction', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 bounds: Tuple[float, float] = None,
                 stoichiometry: Dict['Metabolite', float] = None,
                 gpr: Expression = None,
                 **kwargs):

        """
        Reactions are the primary variables of metabolic models.
        A reaction holds information for bounds, metabolites, stoichiometry and
        gene-protein-reaction rules (the associated between reactions, enzymes/proteins and genes)

        A reaction usually contains the stoichiometry of the metabolites that take part in the reaction.
        A reaction usually contains the GPR rule/logic of the associated genes
        A reaction usually contains the bounds for the fluxes values that can take during simulation

        Dynamic information can be obtained from the previous attributes, such as metabolites, reactants, products,
        genes, boundary, reversibility, charge and mass balance, equation, compartments, etc.

        :param identifier: identifier, e.g. PYK
        :param bounds: tuple having both lower and upper bounds, respectively the minimum and maximum coefficient
        that a reaction can have
        :param stoichiometry: dictionary of metabolites objects and their corresponding stoichiometric value
        :param gpr: an expression object with the boolean logic of the associated genes.
        Genes objects must be associated with the expression object.
        """
        # the coefficient bounds initializer sets minimum and maximum bounds of MEWPY_LB and MEWPY_UB
        if not bounds:
            bounds = ModelConstants.REACTION_LOWER_BOUND, ModelConstants.REACTION_UPPER_BOUND

        elif len(bounds) == 1:
            bounds = (bounds[0], bounds[0])

        elif len(bounds) == 2:
            bounds = tuple(bounds)

        else:
            raise ValueError('Bounds must be a tuple of length 1 or 2')

        if not gpr:
            gpr = Expression()

        if not stoichiometry:
            stoichiometry = {}

        self._bounds = bounds
        self._gpr = Expression()
        self._stoichiometry = {}

        super().__init__(identifier,
                         **kwargs)

        self.replace_stoichiometry(stoichiometry, history=False)

        # The gpr must be an expression object. The expression object properties are used to check whether the genes
        # container is correct.
        self.add_gpr(gpr, history=False)

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------
    @property
    def types(self):

        # noinspection PyUnresolvedReferences
        _types = {Reaction.variable_type}

        _types.update(super(Reaction, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------
    def __str__(self):
        return f'{self.id} || {self.equation}'

    def _reaction_to_html(self):
        """
        It returns a html representation.
        """
        html_dict = {'Equation': self.equation,
                     'Bounds': self.bounds,
                     'Reversibility': self.reversibility,
                     'Metabolites': ', '.join(self.metabolites),
                     'Boundary': self.boundary,
                     'GPR': self.gene_protein_reaction_rule,
                     'Genes': ', '.join(self.genes),
                     'Compartments': ', '.join(self.compartments),
                     'Charge balance': self.charge_balance,
                     'Mass balance': self.mass_balance}

        return html_dict

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------
    @serialize('bounds', 'bounds', '_bounds')
    @property
    def bounds(self) -> Tuple[float, float]:
        """
        Bounds of the reaction. The lower and upper bounds of the reaction
        :return: bounds tuple
        """
        return self._bounds

    @serialize('stoichiometry', 'stoichiometry', '_stoichiometry')
    @property
    def stoichiometry(self) -> Dict['Metabolite', Union[int, float]]:
        """
        Stoichiometry of the reaction. A dictionary of metabolites and their corresponding stoichiometric value
        :return: stoichiometry as a dictionary
        """
        return self._stoichiometry.copy()

    @serialize('gpr', 'gpr', '_gpr')
    @property
    def gpr(self) -> Expression:
        """
        Gene-protein-reaction rule of the reaction. An expression object that holds the boolean logic of the associated
        genes. Genes objects must be associated with the expression object.
        :return: expression object
        """
        return self._gpr

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------
    @bounds.setter
    @recorder
    def bounds(self, value: Union[Tuple[float, float], Tuple[float], float]):
        """
        The setter for the bounds of the reaction.
        It accepts a tuple of lower and upper bounds.
        This setter replaces the current bounds with a new one.
        It also handles the linear problems associated to this reaction.
        :param value: tuple of lower and upper bounds
        :return:
        """
        if value is None:
            value = ModelConstants.REACTION_LOWER_BOUND, ModelConstants.REACTION_UPPER_BOUND

        elif isinstance(value, (int, float)):
            value = (value, value)

        elif len(value) == 1:
            value = (value[0], value[0])

        elif len(value) == 2:
            value = tuple(value)

        else:
            raise ValueError('Invalid value for bounds')

        self._bounds = value

        if self.model:
            self.model.notify()

    @stoichiometry.setter
    @recorder
    def stoichiometry(self, value: Dict['Metabolite', Union[int, float]]):
        """
        The setter for the stoichiometry of the reaction.
        It accepts a dictionary of metabolites and their corresponding stoichiometric value.
        This setter replaces the current stoichiometry with a new one.
        It also handles the linear problems associated to this reaction.
        :param value: stoichiometry as a dictionary
        :return: stoichiometry as a dictionary
        """

        self.replace_stoichiometry(value, history=False)

    @gpr.setter
    @recorder
    def gpr(self, value: Expression):
        """
        The setter for the gpr of the reaction.
        It accepts an expression object that holds the boolean logic of the associated genes.
        Genes objects must be associated with the expression object.
        This setter replaces the current gpr with a new one.
        It also handles the linear problems associated to this reaction.
        :param value: expression object
        :return:
        """

        self.remove_gpr(history=False)
        self.add_gpr(value, history=False)

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------
    @property
    def metabolites(self) -> Dict[str, 'Metabolite']:
        """
        Metabolites of the reaction
        :return: dictionary of metabolites
        """
        return {met.id: met for met in self._stoichiometry}

    @property
    def products(self) -> Dict[str, 'Metabolite']:
        """
        Products of the reaction
        :return: dictionary of products
        """
        return {met.id: met for met, st in self._stoichiometry.items()
                if st > 0.0}

    @property
    def reactants(self) -> Dict[str, 'Metabolite']:
        """
        Reactants of the reaction
        :return: dictionary of reactants
        """
        return {met.id: met for met, st in self._stoichiometry.items()
                if st < 0.0}

    @property
    def compartments(self) -> set:
        """
        Compartments of the reaction
        :return: set of compartments
        """
        return {met.compartment
                for met in self.yield_metabolites()
                if met.compartment is not None}

    @property
    def lower_bound(self) -> Union[int, float]:
        """
        Lower bound of the reaction
        :return: lower bound as a float
        """
        return self.bounds[0]

    @property
    def upper_bound(self) -> Union[int, float]:
        """
        Upper bound of the reaction
        :return: upper bound as a float
        """
        return self.bounds[1]

    @property
    def reversibility(self) -> bool:
        """
        Reversibility of the reaction
        :return: reversibility as a boolean
        """
        return self.lower_bound < - ModelConstants.TOLERANCE and self.upper_bound > ModelConstants.TOLERANCE

    @property
    def equation(self) -> str:
        """
        Equation of the reaction
        :return: equation as a string
        """

        reactants = ' + '.join([f'{abs(st)} {met.id}' for met, st in self._stoichiometry.items() if st < 0.0])
        products = ' + '.join([f'{abs(st)} {met.id}' for met, st in self._stoichiometry.items() if st > 0.0])

        if self.reversibility:
            return f'{reactants} <-> {products}'
        else:
            return f'{reactants} -> {products}'

    @property
    def genes(self) -> Dict[str, 'Gene']:
        """
        Genes of the reaction
        :return: dictionary of genes
        """
        return self.gpr.variables.copy()

    @property
    def gene_protein_reaction_rule(self) -> str:
        """
        Gene-protein-reaction rule of the reaction
        :return: gene-protein-reaction rule as a string
        """
        return self.gpr.to_string()

    @property
    def boundary(self) -> bool:
        """
        Boundary of the reaction
        :return: boundary as a boolean
        """
        return not (self.reactants and self.products)

    @property
    def charge_balance(self) -> Dict[str, Union[int, float]]:
        """
        Charge balance of the reaction
        :return: charge balance as a dictionary
        """
        return {'reactants': sum([self._stoichiometry[met] * met.charge for met in self.yield_reactants()]),
                'products': sum([self._stoichiometry[met] * met.charge for met in self.yield_products()])}

    @property
    def mass_balance(self) -> Dict[str, Union[int, float]]:
        """
        Mass balance of the reaction
        :return: mass balance as a dictionary
        """

        balance = {}

        for metabolite, st in self._stoichiometry.items():

            for atom, count in metabolite.atoms.items():
                amount = count * st

                balance[atom] = balance.get(atom, 0) + amount

        return balance

    # -----------------------------------------------------------------------------
    # Dynamic attributes setters
    # -----------------------------------------------------------------------------
    @lower_bound.setter
    @recorder
    def lower_bound(self, value: Union[float, int]):
        """
        The setter for the lower bound of the reaction.
        It accepts a float or an integer.
        This setter replaces the current lower bound with a new one.
        It also handles the linear problems associated to this reaction.
        :param value: lower bound
        :return:
        """

        if not value:
            value = ModelConstants.REACTION_LOWER_BOUND

        value = (value, self.upper_bound)

        self._bounds = value

        if self.model:
            self.model.notify()

    @upper_bound.setter
    @recorder
    def upper_bound(self, value: Union[float, int]):
        """
        The setter for the upper bound of the reaction.
        It accepts a float or an integer.
        This setter replaces the current upper bound with a new one.
        It also handles the linear problems associated to this reaction.
        :param value: upper bound
        :return:
        """

        if not value:
            value = ModelConstants.REACTION_UPPER_BOUND

        self._bounds = (self.lower_bound, value)

        if self.model:
            self.model.notify()

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------
    def yield_genes(self) -> Generator['Gene', None, None]:
        """
        Generator of genes of the reaction
        :return: generator of genes
        """
        return generator(self.genes)

    def yield_metabolites(self) -> Generator['Metabolite', None, None]:
        """
        Generator of metabolites of the reaction
        :return: generator of metabolites
        """
        return generator(self.metabolites)

    def yield_reactants(self) -> Generator['Metabolite', None, None]:
        """
        Generator of reactants of the reaction
        :return: generator of reactants
        """
        return generator(self.reactants)

    def yield_products(self) -> Generator['Metabolite', None, None]:
        """
        Generator of products of the reaction
        :return: generator of products
        """
        return generator(self.products)

    # -----------------------------------------------------------------------------
    # Polymorphic constructors
    # -----------------------------------------------------------------------------
    @classmethod
    def from_stoichiometry(cls,
                           identifier: Any,
                           stoichiometry: Dict['Metabolite', Union[float, int]],
                           bounds: Tuple[Union[float, int], Union[float, int]] = None,
                           gpr: Expression = None,
                           model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                           **kwargs) -> 'Reaction':
        """
        A reaction is defined by its stoichiometry, establishing the relationship between the metabolites and the
        reaction. The stoichiometry is a dictionary of metabolites and their stoichiometric coefficients.

        A reaction can be created from a stoichiometry dictionary.

        :param identifier: identifier of the reaction
        :param stoichiometry: stoichiometry dictionary
        :param bounds: tuple of lower and upper bounds
        :param gpr: gene-protein-reaction rule
        :param model: model of the reaction
        :param kwargs: additional arguments
        :return: a new reaction object
        """
        instance = cls(identifier=identifier,
                       bounds=bounds,
                       stoichiometry=stoichiometry,
                       gpr=gpr,
                       model=model,
                       **kwargs)

        return instance

    @classmethod
    def from_gpr_string(cls,
                        identifier: Any,
                        rule: str,
                        bounds: Tuple[Union[float, int], Union[float, int]] = None,
                        stoichiometry: Dict['Metabolite', Union[float, int]] = None,
                        model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                        **kwargs) -> 'Reaction':
        """
        A reaction is regularly associated with genes. The GPR rule is a boolean expression that relates the genes
        to the reaction. The GPR rule is a string that can be parsed into a boolean expression.

        A reaction can be created from a gene-protein-reaction rule string.

        :param identifier: identifier of the reaction
        :param rule: gene-protein-reaction rule string
        :param bounds: tuple of lower and upper bounds
        :param stoichiometry: stoichiometry dictionary
        :param model: model of the reaction
        :param kwargs: additional arguments
        :return: a new reaction object
        """
        try:

            symbolic = parse_expression(rule)

            genes = variables_from_symbolic(symbolic=symbolic, types=('gene',), model=model)

            expression = Expression(symbolic=symbolic, variables=genes)

        except SyntaxError as exc:

            expression_warning(f'{rule} cannot be parsed')

            raise exc

        instance = cls(identifier=identifier,
                       bounds=bounds,
                       stoichiometry=stoichiometry,
                       gpr=expression,
                       model=model,
                       **kwargs)

        return instance

    @classmethod
    def from_gpr_expression(cls,
                            identifier: Any,
                            gpr: Expression,
                            bounds: Tuple[Union[float, int], Union[float, int]] = None,
                            stoichiometry: Dict['Metabolite', Union[float, int]] = None,
                            model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                            **kwargs) -> 'Reaction':
        """
        A reaction is regularly associated with genes. The GPR rule is a boolean expression that relates the genes
        to the reaction. The GPR rule is a string that can be parsed into a boolean expression.

        A reaction can be created from a gene-protein-reaction expression.

        :param identifier: identifier of the reaction
        :param gpr: gene-protein-reaction expression
        :param bounds: tuple of lower and upper bounds
        :param stoichiometry: stoichiometry dictionary
        :param model: model of the reaction
        :param kwargs: additional arguments
        :return: a new reaction object
        """
        if not isinstance(gpr, Expression):
            raise TypeError(f'expression must be an {Expression} object')

        instance = cls(identifier=identifier,
                       bounds=bounds,
                       stoichiometry=stoichiometry,
                       gpr=gpr,
                       model=model,
                       **kwargs)

        return instance

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------
    def ko(self, minimum_coefficient: Union[int, float] = 0.0, history=True):
        """
        Knockout the reaction by setting the lower and upper bounds to zero.
        It also handles the linear problems associated to this reaction.

        :param minimum_coefficient: minimum coefficient of the reaction
        :param history: record the operation
        :return:
        """
        old_bounds = tuple(self.bounds)

        self._bounds = (minimum_coefficient, minimum_coefficient)

        if self.model:
            self.model.notify()

        if history:
            self.history.queue_command(undo_func=_bounds_setter,
                                       undo_kwargs={'instance': self,
                                                    'old_bounds': old_bounds},
                                       func=self.ko,
                                       kwargs={'minimum_coefficient': minimum_coefficient,
                                               'history': False})
        return

    def update(self,
               bounds: Tuple[Union[float, int], Union[float, int]] = None,
               stoichiometry: Dict['Metabolite', Union[float, int]] = None,
               gpr: Expression = None,
               **kwargs):
        """
        Update the reaction with new bounds, stoichiometry and gene-protein-reaction rule.

        It also handles the linear problems associated to this reaction.

        Note that, some update operations are not registered in history.
        It is strongly advisable to use update outside history context manager

        :param bounds: tuple of lower and upper bounds
        :param stoichiometry: stoichiometry dictionary
        :param gpr: gene-protein-reaction rule
        :param kwargs: additional arguments
        :return:
        """

        super(Reaction, self).update(**kwargs)

        if bounds is not None:
            self.bounds = bounds

        if gpr is not None:
            self.gpr = gpr

        if stoichiometry is not None:
            self.stoichiometry = stoichiometry

    def replace_stoichiometry(self,
                              stoichiometry: Dict['Metabolite', Union[float, int]],
                              remove_orphans_from_model: bool = True,
                              history=True):
        """
        Replace the stoichiometry of the reaction with a new one.

        It also handles the linear problems associated to this reaction.

        :param stoichiometry: stoichiometry dictionary
        :param remove_orphans_from_model: whether to remove orphan metabolites from the model
        :param history: whether to record the operation
        :return:
        """
        if history:
            self.history.queue_command(undo_func=self.replace_stoichiometry,
                                       undo_kwargs={'stoichiometry': self.stoichiometry,
                                                    'remove_orphans_from_model': remove_orphans_from_model,
                                                    'history': False},
                                       func=self.replace_stoichiometry,
                                       kwargs={'stoichiometry': stoichiometry,
                                               'remove_orphans_from_model': remove_orphans_from_model,
                                               'history': history})

        if not stoichiometry and not self._stoichiometry:
            self._stoichiometry = {}

            return

        self.remove_metabolites(list(self.yield_metabolites()),
                                remove_orphans_from_model=remove_orphans_from_model,
                                history=False)

        self.add_metabolites(stoichiometry,
                             history=False)

    def add_metabolites(self,
                        stoichiometry: Dict['Metabolite', Union[float, int]],
                        history=True):
        """
        Add metabolites to the reaction.
        The stoichiometry dictionary is updated with the new metabolites and coefficients.

        It also handles the linear problems associated to this reaction.
        :param stoichiometry: stoichiometry dictionary
        :param history: whether to record the operation
        :return:
        """
        to_add = []

        for metabolite, coefficient in stoichiometry.items():
            # noinspection PyProtectedMember
            metabolite._reactions[self.id] = self

            self._stoichiometry[metabolite] = coefficient

            to_add.append(metabolite)

        if self.model:
            # the add interface will add all metabolites to the simulators. The simulators will retrieve the new
            # constraints for each metabolite and replace it in the LP
            self.model.add(*to_add, comprehensive=False, history=False)

            self.model.notify()

        if history:
            self.history.queue_command(undo_func=self.remove_metabolites,
                                       undo_kwargs={'metabolites': list(stoichiometry.keys()),
                                                    'remove_orphans_from_model': True,
                                                    'history': False},
                                       func=self.add_metabolites,
                                       kwargs={'stoichiometry': stoichiometry,
                                               'history': history})

    def remove_metabolites(self,
                           metabolites: List['Metabolite'],
                           remove_orphans_from_model: bool = True,
                           history=True):
        """
        Remove metabolites from the reaction.
        Metabolites and their coefficients are removed from the stoichiometry dictionary.

        It also handles the linear problems associated to this reaction.
        :param metabolites: list of metabolites
        :param remove_orphans_from_model: whether to remove orphan metabolites from the model
        :param history: whether to record the operation
        :return:
        """
        if isinstance(metabolites, dict):
            metabolites = list(metabolites.values())

        orphan_mets = []
        old_stoichiometry = {}

        for metabolite in metabolites:

            # noinspection PyProtectedMember
            del metabolite._reactions[self.id]

            # noinspection PyProtectedMember
            if not metabolite._reactions:
                orphan_mets.append(metabolite)

            old_stoichiometry[metabolite] = self._stoichiometry[metabolite]

            del self._stoichiometry[metabolite]

        if self.model:

            # the add interface will add all metabolites to the simulators. The simulators will retrieve the new
            # constraints for each metabolite and replace it in the LP

            if orphan_mets and remove_orphans_from_model:
                self.model.remove(*orphan_mets, remove_orphans=False, history=False)

            self.model.notify()

        if history:
            self.history.queue_command(undo_func=self.add_metabolites,
                                       undo_kwargs={'stoichiometry': old_stoichiometry,
                                                    'history': False},
                                       func=self.remove_metabolites,
                                       kwargs={'metabolites': metabolites,
                                               'remove_orphans_from_model': remove_orphans_from_model,
                                               'history': history})

    def add_gpr(self, gpr: Expression, history=True):
        """
        Add a gene-protein-reaction rule to the reaction.

        It also handles the linear problems associated to this reaction.
        :param gpr: gene-protein-reaction expression object
        :param history: whether to record the operation
        :return:
        """
        if not isinstance(gpr, Expression):
            raise TypeError(f'expression must be an {Expression} object. '
                            f'To set None, provide an empty Expression()')

        if history:
            self.history.queue_command(undo_func=self.remove_gpr,
                                       undo_kwargs={'remove_orphans': True,
                                                    'history': False},
                                       func=self.add_gpr,
                                       kwargs={'gpr': gpr,
                                               'history': history})

        if self.gpr:
            self.remove_gpr(history=False)

        to_add = []

        for gene in gpr.variables.values():
            gene.update(reactions={self.id: self},
                        model=self.model)

            to_add.append(gene)

        self._gpr = gpr

        if self.model:
            self.model.add(*to_add, comprehensive=False, history=False)

            self.model.notify()

    def remove_gpr(self, remove_orphans: bool = True, history=True):
        """
        Remove the gene-protein-reaction rule from the reaction.

        It also handles the linear problems associated to this reaction.
        :param remove_orphans: whether to remove orphan genes from the model
        :param history: whether to record the operation
        :return:
        """
        if history:
            self.history.queue_command(undo_func=self.add_gpr,
                                       undo_kwargs={'gpr': self.gpr,
                                                    'history': False},
                                       func=self.remove_gpr,
                                       kwargs={'remove_orphans': remove_orphans,
                                               'history': history})

        to_remove = []

        for gene in self.yield_genes():

            # noinspection PyProtectedMember
            del gene._reactions[self.id]

            # noinspection PyProtectedMember
            if not gene._reactions:
                to_remove.append(gene)

        self._gpr = Expression()

        if self.model:

            if remove_orphans:
                self.model.remove(*to_remove, remove_orphans=False, history=False)

            self.model.notify()



""" Adapted from Daniel Machado's REFRAMED"""
from enum import Enum
from math import inf
from reframed.core.cbmodel import CBModel
from cobra.core.model import Model

from ..simulation.simulation import Simulator


class VarType(Enum):
    """ Enumeration of possible variable types. """
    BINARY = 'binary'
    INTEGER = 'integer'
    CONTINUOUS = 'continuous'


class Parameter(Enum):
    """ Enumeration of parameters common to all solvers. """
    TIME_LIMIT = 0
    FEASIBILITY_TOL = 1
    INT_FEASIBILITY_TOL = 2
    OPTIMALITY_TOL = 3
    MIP_REL_GAP = 4
    MIP_ABS_GAP = 5
    POOL_SIZE = 6
    POOL_GAP = 7


default_parameters = {
    Parameter.FEASIBILITY_TOL: 1e-9,
    Parameter.OPTIMALITY_TOL: 1e-9,
}


class Solver(object):
    """ Abstract class representing a generic solver.

    All solver interfaces should implement the methods defined in this class.
    """

    def __init__(self, model=None):
        self.problem = None
        self.var_ids = []
        self.constr_ids = []
        self.model = model

    def add_variable(self, var_id, lb=-inf, ub=inf, vartype=VarType.CONTINUOUS, update=True):
        """ Add a variable to the current problem.

        Arguments:
            var_id (str): variable identifier
            lb (float): lower bound
            ub (float): upper bound
            vartype (VarType): variable type (default: CONTINUOUS)
            update (bool): update problem immediately (default: True)
        """

    def set_variable_bounds(self, var_id, lb, ub):
        """Modify a variable bounds

        Args:
            var_id (str): [description]
            lb (float): lower bound
            ub (float): upper bound
        """
        pass

    def add_constraint(self, constr_id, lhs, sense='=', rhs=0, update=True):
        """ Add a constraint to the current problem.

        Arguments:
            constr_id (str): constraint identifier
            lhs (dict): variables and respective coefficients
            sense (str): constraint sense (any of: '<', '=', '>'; default '=')
            rhs (float): right-hand side of equation (default: 0)
            update (bool): update problem immediately (default: True)
        """
        pass

    def remove_variable(self, var_id):
        """ Remove a variable from the current problem.

        Arguments:
            var_id (str): variable identifier
        """
        pass

    def remove_variables(self, var_ids):
        """ Remove variables from the current problem.

        Arguments:
            var_ids (list): variable identifiers
        """
        pass

    def remove_constraint(self, constr_id):
        """ Remove a constraint from the current problem.

        Arguments:
            constr_id (str): constraint identifier
        """
        pass

    def remove_constraints(self, constr_ids):
        """ Remove constraints from the current problem.

        Arguments:
            constr_ids (list): constraint identifiers
        """
        pass

    def list_variables(self):
        """ Get a list of the variable ids defined for the current problem.

        Returns:
            list: variable ids
        """
        return self.var_ids

    def list_constraints(self):
        """ Get a list of the constraint ids defined for the current problem.

        Returns:
            list: constraint ids
        """
        return self.constr_ids

    def update(self):
        """ Update internal structure. Used for efficient lazy updating. """
        pass

    def set_objective(self, linear=None, quadratic=None, minimize=True):
        """ Set a predefined objective for this problem.

        Args:
            linear (str or dict): linear coefficients (or a single variable to optimize)
            quadratic (dict): quadratic coefficients (optional)
            minimize (bool): solve a minimization problem (default: True)

        Notes:
            Setting the objective is optional. It can also be passed directly when calling **solve**.

        """
        pass

    def build_problem(self, model):
        """ Create problem structure for a given model.

        Arguments:
            model
        """

        if isinstance(model, (CBModel, Model)):
            self.__build_problem_model(model)
        elif isinstance(model, Simulator):
            self.__build_problem_simulator(model)
        else:
            raise TypeError

    def __build_problem_model(self, model):
        """ Create a problem for metabolic models (REFRAMED or COBRApy)
        Args:
            model: A metabolic model
        """
        from ..simulation import get_simulator
        sim = get_simulator(model)
        self.__build_problem_simulator(sim)

    def __build_problem_simulator(self, simulator):
        """Create a problem for simulators

        Args:
            simulator: A phenotype simulator
        """
        for r_id in simulator.reactions:
            lb, ub = simulator.get_reaction_bounds(r_id)
            self.add_variable(r_id, lb, ub, update=False)
        self.update()

        table = simulator.metabolite_reaction_lookup()
        for m_id in simulator.metabolites:
            self.add_constraint(m_id, table[m_id], update=False)
        self.update()

    def solve(self, linear=None, quadratic=None, minimize=None, model=None, constraints=None, get_values=True,
              shadow_prices=False, reduced_costs=False, pool_size=0, pool_gap=None):
        """ Solve the optimization problem.

        Arguments:
            linear (dict): linear objective (optional)
            quadratic (dict): quadratic objective (optional)
            minimize (bool): solve a minimization problem (default: True)
            model (CBModel): model (optional, leave blank to reuse previous model structure)
            constraints (dict): additional constraints (optional)
            get_values (bool or list): set to false for speedup if you only care about the objective\
                value (default: True)
            shadow_prices (bool): return shadow prices if available (default: False)
            reduced_costs (bool): return reduced costs if available (default: False)
            pool_size (int): calculate solution pool of given size (only for MILP problems)
            pool_gap (float): maximum relative gap for solutions in pool (optional)

        Returns:
            Solution: solution
        """

        # An exception is raised if the subclass does not implement this method.
        raise Exception('Not implemented for this solver.')

    def get_solution_pool(self, get_values=True):
        """ Return a solution pool for MILP problems.
        Must be called after using solve with pool_size argument > 0.

        Arguments:

            get_values (bool or list): set to false for speedup if you only care about the objective\
                value (default: True)

        Returns:
            list: list of Solution objects

        """
        raise Exception('Not implemented for this solver.')

    def set_parameter(self, parameter, value):
        """ Set a parameter value for this optimization problem

        Arguments:
            parameter (Parameter): parameter type
            value (float): parameter value
        """

        raise Exception('Not implemented for this solver.')

    def set_parameters(self, parameters):
        """ Set values for multiple parameters

        Arguments:
            parameters (dict of Parameter to value): parameter values
        """

        for parameter, value in parameters.items():
            self.set_parameter(parameter, value)

    def set_logging(self, enabled=False):
        """ Enable or disable log output:

        Arguments:
            enabled (bool): turn logging on (default: False)
        """

        raise Exception('Not implemented for this solver.')

    def write_to_file(self, filename):
        """ Write problem to file:

        Arguments:
            filename (str): file path
        """

        raise Exception('Not implemented for this solver.')

    def change_coefficients(self, coefficients):
        """Changes variables coefficients in constraints

        :param coefficients: A list of tuples (constraint name, variable name, new value)
        :type coefficients: list
        """
        raise Exception('Not implemented for this solver.')



from typing import Any, Dict, TYPE_CHECKING, Set, Union, List, Tuple, Generator, Sequence

from mewpy.util.constants import ModelConstants

from mewpy.util.history import recorder

from mewpy.util.utilities import generator
from mewpy.germ.models.serialization import serialize
from .variable import Variable
from .variables_utils import coefficients_setter

if TYPE_CHECKING:
    from .reaction import Reaction


class Gene(Variable, variable_type='gene', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 coefficients: Sequence[float] = None,
                 reactions: Dict[str, 'Reaction'] = None,
                 **kwargs):
        """
        A metabolic gene is regularly associated with metabolic reactions. A metabolic gene is inferred from a metabolic
        model by parsing the Gene-Protein-Reactions associations usually encoded as boolean rules.

        A gene contains multiple coefficients for GPR rule evaluation. It usually takes
        either 0 (inactive) or 1 (active)
        A gene object also holds information regarding the reactions that it is associated with.

        In addition, a metabolic gene can also be represented as a target gene in Metabolic-Regulatory models, as it
        can be inferred from the regulatory rules defining regulatory interactions.

        :param identifier: identifier, e.g. b0001
        :param coefficients: the set of coefficients that this gene can take.
        These coefficients can be expanded later. 0 and 1 are added by default
        :param reactions: the dictionary of reactions to which the gene is associated with
        """
        # the coefficient initializer sets minimum and maximum coefficients of 0.0 and 1.0
        if not coefficients:
            coefficients = (0.0, 1.0)

        else:
            coefficients = tuple(coefficients)

        if not reactions:
            reactions = {}

        self._coefficients = coefficients
        self._reactions = reactions

        super().__init__(identifier, **kwargs)

    @property
    def types(self):
        # noinspection PyUnresolvedReferences
        _types = {Gene.variable_type}

        _types.update(super(Gene, self).types)

        return _types

    def _gene_to_html(self):
        """
        It returns a html dict representation.
        """
        html_dict = {'Coefficients': self.coefficients,
                     'Active': self.is_active,
                     'Reactions': ', '.join(self.reactions)}
        return html_dict

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------
    @serialize('coefficients', 'coefficients', '_coefficients')
    @property
    def coefficients(self) -> Tuple[float, ...]:
        """
        The gene coefficients
        :return: The gene coefficients
        """
        if hasattr(self, '_bounds'):

            # if it is a reaction, bounds must be returned
            return self._bounds

        # if it is a metabolite, the bounds coefficient of the exchange reaction must be returned
        elif hasattr(self, 'exchange_reaction'):

            if hasattr(self.exchange_reaction, '_bounds'):

                # noinspection PyProtectedMember
                return self.exchange_reaction._bounds

        return self._coefficients

    @serialize('reactions', 'reactions', '_reactions')
    @property
    def reactions(self) -> Dict[str, 'Reaction']:
        """
        The reactions associated with this gene
        """
        return self._reactions.copy()

    @property
    def is_active(self):
        """
        It checks whether the gene is active or not
        :return: True if the gene is active, False otherwise
        """
        return max(self.coefficients) > ModelConstants.TOLERANCE

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------
    @coefficients.setter
    @recorder
    def coefficients(self, value: Union[float, Sequence[float]]):
        """
        The gene coefficients setter
        :param value: The gene coefficients
        :return:
        """
        coefficients_setter(self, value)

    @reactions.setter
    def reactions(self, value: Dict[str, 'Reaction']):
        """
        The reactions associated with this gene
        :param value: The reactions associated with this gene
        :return:
        """
        if not value:
            value = {}

        self._reactions = value

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------
    def yield_reactions(self) -> Generator['Reaction', None, None]:
        """
        It yields all reactions
        """
        return generator(self._reactions)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------
    def ko(self, minimum_coefficient: float = 0.0, history=True):
        """
        It performs a gene knock-out. This is accomplished by setting the coefficient object to a tuple of zeros.

        This operation can be reverted using the model history

        :param minimum_coefficient: The minimum coefficient that represents a gene KO. 0.0 is used as default
        :param history: Whether to register this operation in the model history
        :return: None
        """
        old_coef = tuple(self.coefficients)

        coefficients_setter(self, minimum_coefficient)

        if history:
            self.history.queue_command(undo_func=coefficients_setter,
                                       undo_kwargs={'instance': self,
                                                    'value': old_coef},
                                       func=self.ko,
                                       kwargs={'minimum_coefficient': minimum_coefficient,
                                               'history': False})
        return

    def update(self,
               coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]] = None,
               reactions: Dict[str, 'Reaction'] = None,
               **kwargs):
        """
        It performs an update operation to this gene.
        The update operation is similar to a dictionary update.

        Note that, some update operations are not registered in history.
        It is strongly advisable to use update outside history context manager

        :param coefficients: The gene coefficients
        :param reactions: The reactions associated with this gene
        :param kwargs: Other arguments for the base variable, such as identifier, name, etc
        :return:
        """
        super(Gene, self).update(**kwargs)

        if coefficients is not None:
            self.coefficients = coefficients

        if reactions is not None:
            self._reactions.update(reactions)



from typing import Any, TYPE_CHECKING, Dict, Generator, Tuple, Sequence, Union

from mewpy.util.constants import ModelConstants
from mewpy.util.history import recorder
from mewpy.germ.models.serialization import serialize
from mewpy.util.utilities import generator
from .variable import Variable
from .variables_utils import coefficients_setter

if TYPE_CHECKING:
    from .interaction import Interaction
    from .regulator import Regulator


class Target(Variable, variable_type='target', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 coefficients: Sequence[float] = None,
                 interaction: 'Interaction' = None,
                 **kwargs):

        """
        A target gene is associated with a single interaction but can be regulated by multiple regulators.
        The regulatory interaction establishes a relationship between a target gene and regulators.
        A target gene holds the coefficients that it can take. While 0 and 1 are added by default,
        these coefficients can be changed later.

        A target gene can also be represented as a metabolic gene in Metabolic-Regulatory models, as it
        can be inferred from the metabolic rules defining metabolic interactions.
        In regulatory models, a target gene can also be represented as a regulator gene, as it can be inferred
        from the regulatory rules defining regulatory interactions.

        :param identifier: identifier, e.g. b0001
        :param coefficients: the set of coefficients that this target can take.
        These coefficients can be expanded later. 0 and 1 are added by default
        :param interactions: the  interaction to which the target is associated with
        """

        # the coefficient initializer sets minimum and maximum coefficients of 0.0 and 1.0
        if not coefficients:
            coefficients = (0.0, 1.0)
        else:
            coefficients = tuple(coefficients)

        if not interaction:
            interaction = None

        self._coefficients = coefficients
        self._interaction = interaction

        super().__init__(identifier, **kwargs)

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------
    @property
    def types(self):

        # noinspection PyUnresolvedReferences
        _types = {Target.variable_type}

        _types.update(super(Target, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------
    def __str__(self):

        if self.interaction:

            return str(self.interaction)

        return f'{self.id} || {self.coefficients}'

    def _target_to_html(self):
        """
        It returns a html representation.
        """
        html_dict = {'Coefficients': self.coefficients,
                     'Active': self.is_active,
                     'Interaction': self.interaction,
                     'Regulators': ', '.join(self.regulators)}
        return html_dict

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('coefficients', 'coefficients', '_coefficients')
    @property
    def coefficients(self) -> Tuple[float, ...]:
        """
        The coefficients of the target gene.
        :return: the coefficients
        """
        if hasattr(self, '_bounds'):

            # if it is a reaction, bounds must be returned
            return self._bounds

        # if it is a metabolite, the bounds coefficient of the exchange reaction must be returned
        elif hasattr(self, 'exchange_reaction'):

            if hasattr(self.exchange_reaction, '_bounds'):
                # noinspection PyProtectedMember
                return self.exchange_reaction._bounds

        return self._coefficients

    @serialize('interaction', 'interaction', '_interaction')
    @property
    def interaction(self) -> 'Interaction':
        """
        The interaction to which the target is associated with.
        :return: the interaction
        """
        return self._interaction

    @property
    def is_active(self):
        """
        It checks whether the gene is active or not
        :return: True if the gene is active, False otherwise
        """
        return max(self.coefficients) > ModelConstants.TOLERANCE

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------
    @coefficients.setter
    @recorder
    def coefficients(self, value: Union[float, Sequence[float]]):
        """
        The target coefficients setter
        :param value: The target coefficients
        :return:
        """
        coefficients_setter(self, value)

    @interaction.setter
    @recorder
    def interaction(self, value: 'Interaction'):
        """
        The interaction setter.
        It sets the interaction and adds the target to the interaction.

        It also handles the linear problems associated with the interaction.
        :param value: the interaction to which the target is associated with
        :return:
        """
        if not value:
            value = None

        value.add_target(self, history=False)

        self._interaction = value

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------
    @property
    def regulators(self) -> Dict[str, 'Regulator']:
        """
        The regulators that regulate the target gene.
        :return: the regulators as a dictionary
        """
        if self.interaction:
            return self.interaction.regulators.copy()

        return {}

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------
    def yield_regulators(self) -> Generator['Regulator', None, None]:
        """
        A generator that yields the regulators that regulate the target gene.
        :return:
        """
        return generator(self.regulators)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------
    def ko(self, minimum_coefficient: float = 0.0, history=True):
        """
        Knock-out the target gene.
        :param minimum_coefficient: the minimum coefficient
        :param history: whether to record the operation
        :return:
        """
        old_coef = tuple(self.coefficients)

        coefficients_setter(self, minimum_coefficient)

        if history:
            self.history.queue_command(undo_func=coefficients_setter,
                                       undo_kwargs={'instance': self,
                                                    'value': old_coef},
                                       func=self.ko,
                                       kwargs={'minimum_coefficient': minimum_coefficient,
                                               'history': False})

    def update(self,
               coefficients: Sequence[float] = None,
               interaction: 'Interaction' = None,
               **kwargs):
        """
        It updates the target gene with relevant information.

        It also handles the linear problems associated to the interaction.

        Note that, some update operations are not registered in history.
        It is strongly advisable to use update outside history context manager

        :param coefficients: the set of coefficients that this target can take.
        :param interaction: the  interaction to which the target is associated with
        :param kwargs: additional arguments
        :return:
        """
        super(Target, self).update(**kwargs)

        if coefficients is not None:
            self.coefficients = coefficients

        if interaction is not None:
            self.interaction = interaction



import os
from functools import partial
from typing import Union, TYPE_CHECKING

from mewpy.io.dto import DataTransferObject, VariableRecord, History, FunctionTerm, CompartmentRecord
from mewpy.germ.algebra import Expression, Symbol, Or, And, NoneAtom
from mewpy.germ.models import RegulatoryModel, MetabolicModel
from mewpy.util.constants import ModelConstants
from .engine import Engine
from .engines_utils import (build_symbolic,
                            pattern_notes,
                            f_id, F_GENE, F_SPECIE, F_REACTION, F_SPECIE_REV, F_GENE_REV, F_REACTION_REV, convert_fbc,
                            get_sbml_doc_to_write, get_sbml_doc_to_read,
                            UNIT_ID, UNITS,
                            add_sbml_parameter, LOWER_BOUND_ID, UPPER_BOUND_ID, ZERO_BOUND_ID,
                            BOUND_MINUS_INF, BOUND_PLUS_INF,
                            SBO_DEFAULT_FLUX_BOUND, SBO_FLUX_BOUND,
                            get_sbml_lb_id, get_sbml_ub_id, write_sbml_doc,
                            set_gpr,
                            expression_warning, sbml_warning)

if TYPE_CHECKING:
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


class MetabolicSBML(Engine):
    """
    SBML engine for metabolic models.
    Credits to the cobrapy team (https://github.com/opencobra/cobrapy), as this SBML engine is heavily inspired in their
    SBML parser.
    """

    def __init__(self, io, config, model=None):
        super().__init__(io, config, model)

    @property
    def model_type(self):
        return 'metabolic'

    @property
    def model(self):

        if self._model is None:

            identifier = self.dto.id

            if not identifier:
                identifier = self.get_identifier()

            return MetabolicModel(identifier=identifier)

        return self._model

    def parse_notes(self, notes):

        notes_store = {}

        for match in pattern_notes.finditer(notes):

            try:
                key, value = match.group("content").split(":", 1)

            except ValueError:
                self.warnings.append(partial(sbml_warning,
                                             "Unexpected content format {}.".format(match.group("content"))))
                continue

            value = value.strip()

            if value:
                notes_store[key.strip()] = value

        return notes_store

    @staticmethod
    def parse_symbols(fbc_association):

        gene_id = f_id(fbc_association.getGeneProduct(), F_GENE)

        symbol = Symbol(gene_id)

        return symbol

    def parse_leaves(self, fbc_association):

        if fbc_association.isGeneProductRef():

            return self.parse_symbols(fbc_association)

        else:

            return

    def _parse_gpa(self, fbc_association):

        if fbc_association.isFbcOr():

            args = [self._parse_gpa(child)
                    for child in fbc_association.getListOfAssociations()]

            return Or(variables=args)

        elif fbc_association.isFbcAnd():

            args = [self._parse_gpa(child)
                    for child in fbc_association.getListOfAssociations()]

            return And(variables=args)

        else:

            return self.parse_leaves(fbc_association)

    def parse_gpa(self, fbc_association):

        """
        Reads and parses a node of type math ASTNode into a boolean algebraic expression.

        :param fbc_association: FBCAssociation
        :return: Expression
        """

        try:

            symbolic = self._parse_gpa(fbc_association)

        except SyntaxError:

            self.warnings.append(partial(expression_warning,
                                         f'{fbc_association} cannot be parsed. Assigning empty expression instead'))

            symbolic = NoneAtom()

        return symbolic

    def get_identifier(self):

        if os.path.exists(self.io):
            _, identifier = os.path.split(self.io)
            return os.path.splitext(identifier)[0]

        return 'model'

    def _open_to_read(self):

        self._dto = DataTransferObject()

        # -----------------------------------------------------------------------------
        # Doc, Model and FBC Plugin
        # -----------------------------------------------------------------------------

        self.dto.doc = get_sbml_doc_to_read(self.io)

        self.dto.model = self.dto.doc.getModel()

        if self.dto.model is None:
            raise OSError(f'{self.io} is not a valid input. Model SBML section is missing. '
                          f'Provide a correct path or file handler')

        identifier = self.dto.model.getIdAttribute()

        if not identifier:
            self.warnings.append(partial(sbml_warning, 'Model identifier is not encoded in the SBML file'))

            identifier = self.get_identifier()

        self.dto.id = identifier
        self.dto.name = self.dto.model.getName()

    def _open_to_write(self):

        self._dto = DataTransferObject()

        # -----------------------------------------------------------------------------
        # Doc, Model and FBC Plugin
        # -----------------------------------------------------------------------------

        self.dto.doc = get_sbml_doc_to_write(self.io,
                                             level=3,
                                             version=1,
                                             packages=('fbc',),
                                             packages_version=(2,),
                                             packages_required=(False,),
                                             sbo_term=True)

        if self.dto.model is None:

            self.dto.model = self.dto.doc.createModel()

        else:

            self.dto.model = self.dto.doc.getModel()

        # fbc plugin is added by get_sbml_doc_to_write
        self.dto.fbc_plugin = self.dto.model.getPlugin('fbc')
        self.dto.fbc_plugin.setStrict(True)

        if self.model.id is not None:
            self.dto.model.setId(self.model.id)
            self.dto.model.setMetaId('meta_' + self.model.id)

        else:
            self.dto.model.setMetaId('meta_model')

        if self.model.name is not None:
            self.dto.model.setName(self.model.name)

    def open(self, mode='r'):

        if mode == 'r':

            return self._open_to_read()

        elif mode == 'w':

            return self._open_to_write()

        else:
            raise ValueError(f'{mode} mode is not recognized. Try one of the following: r, w')

    def parse(self):

        if self.dto is None:
            raise OSError('SBML file is not open')

        if self.dto.id is None:
            raise OSError('SBML file is not open')

        if self.dto.doc is None:
            raise OSError('SBML file is not open')

        if self.dto.model is None:
            raise OSError(f'SBML file is not open')

        self.dto.fbc_plugin = self.dto.model.getPlugin("fbc")

        if not self.dto.fbc_plugin:
            self.warnings.append(partial(sbml_warning, "SBML model does not have fbc plugin"))

        else:
            if not self.dto.fbc_plugin.isSetStrict():
                self.warnings.append(partial(sbml_warning, 'SBML model fbc plugin is not set to strict. It must '
                                                           'fbc:strict="true"'))

            doc_fbc = self.dto.doc.getPlugin("fbc")
            fbc_version = doc_fbc.getPackageVersion()

            # fbc 1 to 2. If fails, an import error is launched
            if fbc_version == 1:
                self.warnings.append(partial(sbml_warning, 'Models should be encoded using fbc version 2. Converting '
                                                           'fbc v1 to fbc v2'))

                convert_fbc(self.dto.doc)

        # -----------------------------------------------------------------------------
        # Model id, name, etc
        # -----------------------------------------------------------------------------

        self.dto.level = self.dto.model.getLevel()
        self.dto.version = self.dto.model.getVersion()

        has_history = self.dto.model.isSetModelHistory()

        if has_history:

            history = self.dto.model.getModelHistory()

            created = None
            if history.isSetCreatedDate():
                created = history.getCreatedDate()

            creators = history.getListCreators()

            self.dto.history = History(created, creators)

        # -----------------------------------------------------------------------------
        # Compartments
        # -----------------------------------------------------------------------------

        for compartment in self.dto.model.getListOfCompartments():
            self.dto.compartments[compartment.getIdAttribute()] = CompartmentRecord(id=compartment.getIdAttribute(),
                                                                                    name=compartment.getName())

        # -----------------------------------------------------------------------------
        # Metabolites and Extracellular metabolites
        # -----------------------------------------------------------------------------

        metabolites = {}
        extracellular_metabolites = {}

        if self.dto.model.getNumSpecies() == 0:
            self.warnings.append(partial(sbml_warning, "SBML model does not have species/metabolites"))

        for met in self.dto.model.getListOfSpecies():

            met_id = f_id(met.getIdAttribute(), F_SPECIE)
            met_name = met.getName()
            met_aliases = {met_id, met_name}

            if not met_name:
                met_name = met_id

            met_notes = met.getNotesString()
            met_annotation = met.getAnnotationString()
            met_compartment = met.getCompartment()

            met_fbc = met.getPlugin("fbc")
            if met_fbc:
                met_charge = met_fbc.getCharge()
                met_formula = met_fbc.getChemicalFormula()
            else:
                if met.isSetCharge():
                    met_charge = met.getCharge()
                    met_formula = None
                else:
                    met_charge = None
                    met_formula = None

            met_record = VariableRecord(id=met_id,
                                        name=met_name,
                                        aliases=met_aliases,
                                        notes=met_notes,
                                        annotation=met_annotation,
                                        compartment=met_compartment,
                                        charge=met_charge,
                                        formula=met_formula)

            self.variables[met_id].add('metabolite')

            if met.getBoundaryCondition() is True:
                # extracellular metabolites
                extracellular_metabolites[met_id] = met_record

            metabolites[met_id] = met_record

            self.dto.variables[met_id] = met_record

        self.dto.metabolites = metabolites
        self.dto.extracellular_metabolites = extracellular_metabolites

        # -----------------------------------------------------------------------------
        # Genes
        # -----------------------------------------------------------------------------

        genes = {}

        # parsing genes encoded in the gene products section of the sbml
        if self.dto.fbc_plugin:

            if self.dto.fbc_plugin.getNumGeneProducts() == 0:
                self.warnings.append(partial(sbml_warning, "SBML model fbc plugin does not have gene products/genes"))

            for gene in self.dto.fbc_plugin.getListOfGeneProducts():

                gene_id = f_id(gene.getIdAttribute(), F_GENE)
                gene_name = gene.getName()

                if not gene_name:
                    gene_name = gene_id

                gene_aliases = {gene_id, gene_name}

                gene_notes = gene.getNotesString()
                gene_annotation = gene.getAnnotationString()

                gene_record = VariableRecord(id=gene_id,
                                             name=gene_name,
                                             aliases=gene_aliases,
                                             notes=gene_notes,
                                             annotation=gene_annotation)

                self.variables[gene_id].add('gene')

                genes[gene_id] = gene_record
                self.dto.variables[gene_id] = gene_record

        self.dto.genes = genes

        # -----------------------------------------------------------------------------
        # Reactions
        # -----------------------------------------------------------------------------

        reactions = {}

        if self.dto.model.getNumReactions() == 0:
            self.warnings.append(partial(sbml_warning, "SBML model does not have reactions"))

        for reaction in self.dto.model.getListOfReactions():

            rxn_id = f_id(reaction.getIdAttribute(), F_REACTION)
            rxn_name = reaction.getName()

            if not rxn_name:
                rxn_name = rxn_id

            rxn_aliases = {rxn_id, rxn_name}

            rxn_notes = self.parse_notes(reaction.getNotesString())
            rxn_annotation = reaction.getAnnotationString()

            # ------------------------------------------------
            # Reaction fbc plugin is of the upmost importance to parse reaction properties
            # ------------------------------------------------

            rxn_fbc = reaction.getPlugin("fbc")

            # ------------------------------------------------
            # Reaction Bounds
            # ------------------------------------------------

            if rxn_fbc:

                rxn_bounds = [ModelConstants.REACTION_LOWER_BOUND, ModelConstants.REACTION_UPPER_BOUND]

                # bounds in fbc parameters section
                bounds_ids = (rxn_fbc.getLowerFluxBound(), rxn_fbc.getUpperFluxBound())

                for i, bound_id in enumerate(bounds_ids):

                    if bound_id:

                        bound_parameter = self.dto.model.getParameter(bound_id)

                        if bound_parameter and bound_parameter.getConstant() and bound_parameter.getValue() is not None:

                            rxn_bounds[i] = bound_parameter.getValue()

                        else:

                            self.warnings.append(
                                partial(sbml_warning, f"Incorrect {bound_parameter} bound for {reaction} reaction. "
                                                      f"Set to default"))

                    else:

                        self.warnings.append(partial(sbml_warning, f"Bound for {reaction} reaction not found in the "
                                                                   f"SBML model fbc plugin. Set to "
                                                                   f"default. Try to set all bounds explicitly on all "
                                                                   f"reactions"))

            elif reaction.isSetKineticLaw():

                self.warnings.append(partial(sbml_warning, f"{reaction} reaction fbc plugin not found. This might "
                                                           f"hinder reaction parsing"))

                self.warnings.append(partial(
                    sbml_warning, f"Bounds have been detected in kinetic laws for {reaction} reaction. "
                                  f"Try to set all bounds explicitly on all reactions using the fbc plugin, as mewpy "
                                  f"can miss sometimes kinetic laws"))

                # bounds encoded in the kinetic law. Not advised
                kinetic_law = reaction.getKineticLaw()

                rxn_bounds = [ModelConstants.REACTION_LOWER_BOUND, ModelConstants.REACTION_UPPER_BOUND]

                # parameters of the kinetic law
                kinetic_parameters = ('LOWER_BOUND', 'UPPER_BOUND')

                for i, kinetic_parameter in enumerate(kinetic_parameters):

                    bound = kinetic_law.getParameter(kinetic_parameter)

                    if bound:
                        rxn_bounds[i] = bound.getValue()

                    else:
                        self.warnings.append(partial(sbml_warning,
                                                     f"{kinetic_parameter} has not been detected. "
                                                     f"{kinetic_parameter} has been set to default. "
                                                     f"Try to set all bounds explicitly on all reactions using "
                                                     f"the fbc plugin"))

            else:

                self.warnings.append(partial(sbml_warning, f"{reaction} reaction fbc plugin not found. This might "
                                                           f"hinder reaction parsing"))

                self.warnings.append(partial(sbml_warning,
                                             f"Bounds have not been detected. Bounds have been set to default. Try to "
                                             f"set all bounds explicitly on all reactions using the fbc plugin"))

                rxn_bounds = [ModelConstants.REACTION_LOWER_BOUND, ModelConstants.REACTION_UPPER_BOUND]

            # ------------------------------------------------
            # Reaction stoichiometry and metabolites
            # ------------------------------------------------

            rxn_stoichiometry = {}
            rxn_metabolites = {}
            rxn_products = {}
            rxn_reactants = {}

            # reactants
            for reactant in reaction.getListOfReactants():

                reactant_id = f_id(reactant.getSpecies(), F_SPECIE)

                if reactant_id not in self.dto.metabolites:
                    raise ValueError(f"{reactant_id} reactant of {reaction} reaction "
                                     f"is not listed as specie in the SBML Model.")

                reactant_record = self.dto.metabolites[reactant_id]

                rxn_metabolites[reactant_id] = reactant_record
                rxn_reactants[reactant_id] = reactant_record
                rxn_stoichiometry[reactant_id] = -reactant.getStoichiometry()

            # products
            for product in reaction.getListOfProducts():

                product_id = f_id(product.getSpecies(), F_SPECIE)

                if product_id not in self.dto.metabolites:
                    raise ValueError(f"{product_id} product of {reaction} reaction "
                                     f"is not listed as specie in the SBML Model.")

                product_record = self.dto.metabolites[product_id]

                rxn_metabolites[product_id] = product_record
                rxn_products[product_id] = product_record
                rxn_stoichiometry[product_id] = product.getStoichiometry()

            # ------------------------------------------------
            # Reaction Gene-Protein-Reaction rule
            # ------------------------------------------------

            symbolic = NoneAtom()

            if rxn_fbc:

                # if the gpr is encoded in the reaction fbc plugin.
                # It is advised.

                gpa = rxn_fbc.getGeneProductAssociation()

                if gpa is not None:
                    symbolic = self.parse_gpa(gpa.getAssociation())

            else:

                self.warnings.append(partial(sbml_warning,
                                             "Please use fbc plugin fbc:gpr to encode gprs in the future, "
                                             "as parsing gprs from notes might be troublesome"))

                # Else the gpr parsing tries to find the gpr rule (string) within the notes

                gpr_rule = rxn_notes.get('GENE ASSOCIATION',
                                         rxn_notes.get('GENE_ASSOCIATION', None))

                if gpr_rule is None:

                    self.warnings.append(partial(sbml_warning,
                                                 "GPR was not found within the reaction's notes section"))

                else:

                    gpr_rule = ' '.join(f_id(child, F_GENE) for child in gpr_rule.split(' '))

                    symbolic, warning = build_symbolic(expression=gpr_rule)

                    if warning:
                        self.warnings.append(partial(expression_warning, warning))

            rxn_genes = {}
            for symbol in symbolic.atoms(symbols_only=True):

                if rxn_fbc:

                    if symbol.name in self.dto.genes:

                        rxn_genes[symbol.name] = self.dto.genes[symbol.name]

                        continue

                    else:

                        self.warnings.append(partial(sbml_warning,
                                                     f'{symbol.name} is not listed in the SBML model fbc plugin'))

                gene_record = VariableRecord(id=symbol.name,
                                             name=symbol.name,
                                             aliases={symbol.name, symbol.value})

                self.variables[symbol.name].add('gene')

                self.dto.variables[symbol.name] = gene_record
                self.dto.genes[symbol.name] = gene_record

                rxn_genes[symbol.name] = gene_record

            # -----------------------------------------------------------------------------
            # GPR Function term
            # -----------------------------------------------------------------------------
            function_term = FunctionTerm(id='gpr_term', symbolic=symbolic, coefficient=1)

            # ------------------------------------------------
            # Building reaction record and assigning to containers
            # ------------------------------------------------

            reaction_record = VariableRecord(id=rxn_id,
                                             name=rxn_name,
                                             aliases=rxn_aliases,
                                             notes=rxn_notes,
                                             annotation=rxn_annotation,
                                             bounds=tuple(rxn_bounds),
                                             genes=rxn_genes,
                                             gpr=function_term,
                                             metabolites=rxn_metabolites,
                                             products=rxn_products,
                                             reactants=rxn_reactants,
                                             stoichiometry=rxn_stoichiometry)

            self.variables[rxn_id].add('reaction')

            reactions[rxn_id] = reaction_record

            self.dto.variables[rxn_id] = reaction_record

        self.dto.reactions = reactions

        # Some transport reactions have fictitious extracellular metabolites, encoded with the prefix _b,
        # such as M_a_b. In this case, a exchange reaction is created for each metabolite

        extracellular_reactions = {}

        for extracellular_met in self.dto.extracellular_metabolites.values():
            # ------------------------------------------------
            # Building reaction record and assigning to containers
            # ------------------------------------------------

            reaction_record = VariableRecord(id=f'EX_{extracellular_met.id}',
                                             name=f'EX_{extracellular_met.id}',
                                             bounds=(ModelConstants.REACTION_LOWER_BOUND,
                                                     ModelConstants.REACTION_UPPER_BOUND),
                                             metabolites={extracellular_met.id: extracellular_met},
                                             reactants={extracellular_met.id: extracellular_met},
                                             stoichiometry={extracellular_met.id: -1})

            self.variables[f'EX_{extracellular_met.id}'].add('reaction')

            extracellular_reactions[f'EX_{extracellular_met.id}'] = reaction_record

            self.dto.variables[f'EX_{extracellular_met.id}'] = reaction_record

            self.warnings.append(partial(sbml_warning,
                                         f'EX_{extracellular_met.id} reaction added '
                                         f'for metabolite {extracellular_met.id}'))

        self.dto.extracellular_reactions = extracellular_reactions
        self.dto.reactions.update(extracellular_reactions)

        # -----------------------------------------------------------------------------
        # Objective
        # -----------------------------------------------------------------------------

        model_objective = {}

        if self.dto.fbc_plugin:

            objectives = self.dto.fbc_plugin.getListOfObjectives()

            if objectives is None:
                self.warnings.append(partial(sbml_warning, "listOfObjectives element not found"))

            elif objectives.size() == 0:
                self.warnings.append(partial(sbml_warning, "No objective in listOfObjectives"))

            elif not objectives.getActiveObjective():
                self.warnings.append(partial(sbml_warning, "No active objective in listOfObjectives"))

            else:
                active_objectives = objectives.getActiveObjective()
                objective = self.dto.fbc_plugin.getObjective(active_objectives)

                direction = objective.getType()

                for flux_objective in objective.getListOfFluxObjectives():

                    flux_objective_id = f_id(flux_objective.getReaction(), F_REACTION)

                    objective_rxn = self.dto.reactions.get(flux_objective_id, None)

                    if objective_rxn is None:
                        self.warnings.append(partial(sbml_warning,
                                                     f"Objective {flux_objective_id} reaction "
                                                     f"not found in the SBML model"))

                        continue

                    coef = flux_objective.getCoefficient()

                    if direction == 'minimize':
                        coef = -coef

                    model_objective[flux_objective_id] = coef

        else:
            self.warnings.append(partial(sbml_warning,
                                         f"Objective might be encoded in kinetic laws of a given reaction. However, "
                                         f"mewpy does not handle kinetic laws. The objective has not been set. Try "
                                         f"to set the objective explicitly on the fbc plugin"))

        if len(model_objective) == 0:
            self.warnings.append(partial(sbml_warning, "No objective found for the model"))

        self.dto.objective = model_objective

    def read(self,
             model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
             variables=None):

        if not model:
            model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = self.model

        if not variables:
            variables = self.variables

        if self.dto.id:
            model._id = self.dto.id

        if self.dto.name:
            model.name = self.dto.name

        model.compartments = {compartment.id: compartment.name
                              for compartment in self.dto.compartments.values()}

        processed_metabolites = set()
        processed_genes = set()

        for rxn_id, rxn_record in self.dto.reactions.items():

            genes = {}

            for gene_id, gene_record in rxn_record.genes.items():

                gene, warning = gene_record.to_variable(model=model,
                                                        types=variables.get(gene_id, {'gene'}),
                                                        name=gene_record.name,
                                                        aliases=gene_record.aliases)

                if warning:
                    self.warnings.append(partial(sbml_warning, warning))

                genes[gene_id] = gene

                processed_genes.add(gene_id)

            stoichiometry = {}

            for met_id, met_record in rxn_record.metabolites.items():

                met, warning = met_record.to_variable(model=model,
                                                      types=variables.get(met_id, {'metabolite'}),
                                                      name=met_record.name,
                                                      aliases=met_record.aliases,
                                                      compartment=met_record.compartment,
                                                      charge=met_record.charge,
                                                      formula=met_record.formula)

                if warning:
                    self.warnings.append(partial(sbml_warning, warning))

                coef = rxn_record.stoichiometry[met_id]

                stoichiometry[met] = coef

                processed_metabolites.add(met_id)

            gpr = Expression(symbolic=rxn_record.gpr.symbolic, variables=genes)

            rxn, warning = rxn_record.to_variable(model=model,
                                                  types=variables.get(rxn_id, {'reaction'}),
                                                  bounds=rxn_record.bounds,
                                                  gpr=gpr,
                                                  stoichiometry=stoichiometry)

            if warning:
                self.warnings.append(partial(sbml_warning, warning))

            model.add(rxn)

        to_append = []

        for met_id, met_record in self.dto.metabolites.items():

            if met_id not in processed_metabolites:
                met, warning = met_record.to_variable(model=model,
                                                      types=variables.get(met_id, {'metabolite'}),
                                                      name=met_record.name,
                                                      aliases=met_record.aliases,
                                                      compartment=met_record.compartment,
                                                      charge=met_record.charge,
                                                      formula=met_record.formula)

                if warning:
                    self.warnings.append(partial(sbml_warning, warning))

                to_append.append(met)

        for gene_id, gene_record in self.dto.genes.items():

            if gene_id not in processed_genes:
                gene, warning = gene_record.to_variable(model=model,
                                                        types=variables.get(gene_id, {'gene'}),
                                                        name=gene_record.name,
                                                        aliases=gene_record.aliases)

                if warning:
                    self.warnings.append(partial(sbml_warning, warning))

                to_append.append(gene)

        model.add(*to_append)

        model.objective = self.dto.objective

        return model

    def write(self):

        if self.dto is None:
            raise OSError('SBML file is not open')

        if self.dto.doc is None:
            raise OSError('SBML file is not open')

        if self.dto.model is None:
            raise OSError(f'SBML file is not open')

        # -----------------------------------------------------------------------------
        # Units
        # -----------------------------------------------------------------------------
        units = self.config.get('units', False)
        unit_definition = None
        if units:

            unit_definition = self.dto.model.createUnitDefinition()
            unit_definition.setId(UNIT_ID)

            for unit in UNITS:
                _unit = unit_definition.createUnit()
                _unit.setKind(unit.kind)
                _unit.setExponent(unit.exponent)
                _unit.setScale(unit.scale)
                _unit.setMultiplier(unit.multiplier)

        # -----------------------------------------------------------------------------
        # Constants and parameters
        # -----------------------------------------------------------------------------
        add_sbml_parameter(sbml_model=self.dto.model,
                           parameter_id=LOWER_BOUND_ID,
                           value=ModelConstants.REACTION_LOWER_BOUND,
                           constant=True,
                           sbo=SBO_DEFAULT_FLUX_BOUND)

        add_sbml_parameter(sbml_model=self.dto.model,
                           parameter_id=UPPER_BOUND_ID,
                           value=ModelConstants.REACTION_UPPER_BOUND,
                           constant=True,
                           sbo=SBO_DEFAULT_FLUX_BOUND)

        add_sbml_parameter(sbml_model=self.dto.model,
                           parameter_id=ZERO_BOUND_ID,
                           value=0,
                           constant=True,
                           sbo=SBO_DEFAULT_FLUX_BOUND)

        add_sbml_parameter(sbml_model=self.dto.model,
                           parameter_id=BOUND_MINUS_INF,
                           value=-float("Inf"),
                           constant=True,
                           sbo=SBO_DEFAULT_FLUX_BOUND)

        add_sbml_parameter(sbml_model=self.dto.model,
                           parameter_id=BOUND_PLUS_INF,
                           value=float("Inf"),
                           constant=True,
                           sbo=SBO_DEFAULT_FLUX_BOUND)

        # -----------------------------------------------------------------------------
        # Compartments
        # -----------------------------------------------------------------------------
        for compartment_id, compartment_name in self.model.compartments.items():
            sbml_compartment = self.dto.model.createCompartment()
            sbml_compartment.setId(compartment_id)
            sbml_compartment.setName(compartment_name)
            sbml_compartment.setConstant(True)

        # -----------------------------------------------------------------------------
        # Metabolites
        # -----------------------------------------------------------------------------
        for metabolite in self.model.yield_metabolites():
            sbml_specie = self.dto.model.createSpecies()

            sbml_specie.setId(f_id(metabolite.id, F_SPECIE_REV))
            sbml_specie.setConstant(False)
            sbml_specie.setBoundaryCondition(False)
            sbml_specie.setHasOnlySubstanceUnits(False)
            sbml_specie.setName(metabolite.name)
            sbml_specie.setCompartment(metabolite.compartment)
            specie_fbc = sbml_specie.getPlugin('fbc')
            specie_fbc.setCharge(metabolite.charge)
            specie_fbc.setChemicalFormula(metabolite.formula)

        # -----------------------------------------------------------------------------
        # Genes
        # -----------------------------------------------------------------------------
        for gene in self.model.yield_genes():
            gene_product = self.dto.fbc_plugin.createGeneProduct()

            gene_product.setId(f_id(gene.id, F_GENE_REV))
            gene_product.setName(gene.name)
            gene_product.setLabel(gene.id)

        # -----------------------------------------------------------------------------
        # Objective
        # -----------------------------------------------------------------------------
        objective = self.dto.fbc_plugin.createObjective()
        objective.setId('obj')
        objective.setType('maximize')
        self.dto.fbc_plugin.setActiveObjectiveId('obj')

        for reaction, coefficient in self.model.objective.items():
            flux_objective = objective.createFluxObjective()
            flux_objective.setReaction(f_id(reaction.id, F_REACTION_REV))
            flux_objective.setCoefficient(coefficient)

        # -----------------------------------------------------------------------------
        # Reactions
        # -----------------------------------------------------------------------------
        for reaction in self.model.yield_reactions():

            sbml_reaction = self.dto.model.createReaction()
            sbml_reaction.setId(f_id(reaction.id, F_REACTION_REV))
            sbml_reaction.setName(reaction.name)
            sbml_reaction.setFast(False)
            sbml_reaction.setReversible(reaction.reversibility)

            # -----------------------------------------------------------------------------
            # Stoichiometry
            # -----------------------------------------------------------------------------
            for metabolite, st in reaction.stoichiometry.items():

                if st < 0:
                    sbml_reactant = sbml_reaction.createReactant()
                    sbml_reactant.setSpecies(f_id(metabolite.id, F_SPECIE_REV))
                    sbml_reactant.setStoichiometry(-st)
                    sbml_reactant.setConstant(True)

                else:
                    sbml_product = sbml_reaction.createProduct()
                    sbml_product.setSpecies(f_id(metabolite.id, F_SPECIE_REV))
                    sbml_product.setStoichiometry(st)
                    sbml_product.setConstant(True)

            # -----------------------------------------------------------------------------
            # Bounds
            # -----------------------------------------------------------------------------
            sbml_rxn_fbc = sbml_reaction.getPlugin('fbc')

            lb_parameter_id = get_sbml_lb_id(sbml_model=self.dto.model,
                                             reaction=reaction,
                                             unit_definition=unit_definition)

            if lb_parameter_id is None:
                lb_parameter_id = f'{f_id(reaction.id, F_REACTION_REV)}_lb'

                add_sbml_parameter(sbml_model=self.dto.model,
                                   parameter_id=lb_parameter_id,
                                   value=reaction.lower_bound,
                                   sbo=SBO_FLUX_BOUND,
                                   constant=True,
                                   unit_definition=unit_definition)

            sbml_rxn_fbc.setLowerFluxBound(lb_parameter_id)

            ub_parameter_id = get_sbml_ub_id(sbml_model=self.dto.model,
                                             reaction=reaction,
                                             unit_definition=unit_definition)

            if ub_parameter_id is None:
                ub_parameter_id = f'{f_id(reaction.id, F_REACTION_REV)}_lb'

                add_sbml_parameter(sbml_model=self.dto.model,
                                   parameter_id=ub_parameter_id,
                                   value=reaction.upper_bound,
                                   sbo=SBO_FLUX_BOUND,
                                   constant=True,
                                   unit_definition=unit_definition)

            sbml_rxn_fbc.setUpperFluxBound(ub_parameter_id)

            # -----------------------------------------------------------------------------
            # Reaction Gene-Protein-Reaction rule
            # -----------------------------------------------------------------------------
            set_gpr(self, sbml_warning, reaction, sbml_rxn_fbc)

        write_sbml_doc(self.io, self.dto.doc)

    def close(self):

        if hasattr(self.io, 'close'):
            self.io.close()

    def clean(self):
        self._dto = None
